
# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the Economics entity.
  Each component (or source?) can have one of these to describe its economics.
"""
from __future__ import unicode_literals, print_function
import sys
from collections import defaultdict
import numpy as np
from HERON.src import ValuedParams
from HERON.src.ValuedParamHandler import ValuedParamHandler
import HERON.src._utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  framework_path = hutils.get_raven_loc()
  sys.path.append(framework_path)
from ravenframework.utils import InputData, xmlUtils,InputTypes


class CashFlowUser:
  """
    Base class for objects that want to access the functionality of the CashFlow objects.
    Generally this means the CashFlowUser will have an "economics" xml node used to define it,
    and will have a group of cash flows associated with it (e.g. a "component")

    In almost all cases, initialization methods should be called as part of the inheritor's method call.
  """
  @classmethod
  def get_input_specs(cls, spec):
    """
      Collects input specifications for this class.
      Note this needs to be called as part of an inheriting class's specification definition
      @ In, spec, InputData, specifications that need cash flow added to it
      @ Out, spec, InputData, specs
    """
    # this unit probably has some economics
    spec.addSub(CashFlowGroup.get_input_specs())
    return spec

  def __init__(self):
    """
      Constructor
      @ In, None
      @ Out, None
    """
    self._economics = None # CashFlowGroup

  def read_input(self, specs):
    """
      Sets settings from input file
      @ In, specs, InputData params, input from user
      @ Out, None
    """
    self._economics = CashFlowGroup(self)
    self._economics.read_input(specs)

  def get_cashflows(self):
    """
      Getter.
      @ In, None
      @ Out, cashflow, list, cash flows for this cashflow user (ordered)
    """
    return self._economics.get_cashflows()

  def get_crossrefs(self):
    """
      Collect the required value entities needed for this component to function.
      @ In, None
      @ Out, crossrefs, dict, mapping of dictionaries with information about the entities required.
    """
    return self._economics.get_crossrefs()

  def set_crossrefs(self, refs):
    """
      Connect cross-reference material from other entities to the ValuedParams in this component.
      @ In, refs, dict, dictionary of entity information
      @ Out, None
    """
    self._economics.set_crossrefs(refs)

  def get_state_cost(self, activity, meta, marginal=False):
    """
      get the cost given particular activities (state) of the cash flow user
      @ In, raven_vars, dict, additional variables (presumably from raven) that might be needed
      @ In, meta, dict, further dictionary of information that might be needed
      @ In, marginal, bool, optional, if True then only get marginal costs
      @ Out, cost, dict, cost of activity as a breakdown
    """
    return self.get_economics().evaluate_cfs(activity, meta, marginal=marginal)

  def get_economics(self):
    """
      Accessor for economics.
      @ In, None
      @ Out, econ, CashFlowGroup, cash flows for this cash flow user
    """
    return self._economics


class CashFlowGroup:
  """
    Just a holder for multiple cash flows, and methods for doing stuff with them
  """
  ##################
  # INITIALIZATION #
  ##################
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    specs = InputData.parameterInputFactory('economics', ordered=False, baseNode=None,
        descr=r"""this node is where all the economic information about this
              component is placed.""")
    specs.addSub(InputData.parameterInputFactory('lifetime', contentType=InputTypes.IntegerType,
        descr=r"""indicates the number of \emph{cycles} (often \emph{years}) this unit is expected
              to operate before replacement. Replacement is represented as overnight capital cost
              in the year the component is replaced."""))
    cf = CashFlow.get_input_specs()
    specs.addSub(cf)
    return specs

  def __init__(self, component):
    """
      Constructor.
      @ In, component, CashFlowUser instance, object to which this group belongs
      @ Out, None
    """
    self.name = component.name
    self._component = component # component this one
    self._lifetime = None # lifetime of the component
    self._cash_flows = []

  def read_input(self, source, xml=False):
    """
      Sets settings from input file
      @ In, source, InputData.ParameterInput, input from user
      @ In, xml, bool, if True then XML is passed in, not input data
      @ Out, None
    """
    # allow read_input argument to be either xml or input specs
    if xml:
      specs = self.get_input_specs()()
      specs.parseNode(source)
    else:
      specs = source
    # read in specs
    for item in specs.subparts:
      if item.getName() == 'lifetime':
        self._lifetime = item.value
      elif item.getName() == 'CashFlow':
        new = CashFlow(component=self._component)
        new.read_input(item)
        self._cash_flows.append(new)

  def get_crossrefs(self):
    """
      Provides a dictionary of the entities needed by this cashflow group to be evaluated
      @ In, None
      @ Out, crossrefs, dict, dictionary of crossreferences needed (see ValuedParams)
    """
    crossrefs = dict((cf, cf.get_crossrefs()) for cf in self._cash_flows)
    return crossrefs

  def set_crossrefs(self, refs):
    """
      Provides links to entities needed to evaluate this cash flow group.
      @ In, refs, dict, reference entities
      @ Out, None
    """
    # set up pointers
    for cf in list(refs.keys()):
      for try_match in self._cash_flows:
        if try_match == cf:
          try_match.set_crossrefs(refs.pop(try_match))
          break
      else:
        cf.set_crossrefs({})
    # perform checks


  #######
  # API #
  #######
  def evaluate_cfs(self, activity, meta, marginal=False):
    """
      Calculates the incremental cost of a particular system configuration.
      @ In, activity, XArray.DataArray, array of driver-centric variable values
      @ In, meta, dict, additional user-defined meta
      @ In, marginal, bool, optional, if True then only get marginal cashflows (e.g. recurring hourly)
      @ Out, cost, dict, cash flow evaluations
    """
    # combine all cash flows into single cash flow evaluation
    if marginal:
      # FIXME assuming 'year' is the only non-marginal value
      # FIXME why is it "repeating" and not "Recurring"?
      cost = dict((cf.name, cf.evaluate_cost(activity, meta))
                    for cf in self.get_cashflows()
                    if (cf._type == 'repeating' and cf.get_period() != 'year'))
    else:
      cost = dict((cf.name, cf.evaluate_cost(activity, meta))
                    for cf in self.get_cashflows())
    return cost

  def get_cashflows(self):
    """
      Getter.
      @ In, None
      @ Out, cashflow, list, cash flows for this cashflow group (ordered)
    """
    return self._cash_flows

  def get_component(self):
    """
      Return the cash flow user that owns this group
      @ In, None
      @ Out, component, CashFlowUser instance, owner
    """
    return self._component

  def get_lifetime(self):
    """
      Provides the lifetime of this cash flow user.
      @ In, None
      @ Out, lifetime, int, lifetime
    """
    return self._lifetime

  def check_if_finalized(self):
    """
      Check finalization status of cashflows for this group.
      @ In, None
      @ Out, finalized, bool, True if all are finalized
    """
    return all(k.is_finalized() for k in self._cash_flows)

  def finalize(self, activity, raven_vars, meta, times=None):
    """
      Evaluate the parameters for member cash flows, and freeze values so they aren't changed again.
      @ In, activity, dict, mapping of variables to values (may be np.arrays)
      @ In, raven_vars, dict, TODO part of meta! Consolidate!
      @ In, times, list, optional, times to finalize values for
      @ Out, None
    """
    info = {'raven_vars': raven_vars, 'meta': meta}
    for cf in self._cash_flows:
      cf.finalize(activity, info, times=times)

  def calculate_lifetime_cashflows(self):
    """
      Passthrough to CashFlow method of the same name.
      @ In, None
      @ Out, None
    """
    for cf in self._cash_flows:
      cf.calculate_lifetime_cashflow(self._lifetime)





class CashFlow:
  """
    Hold the economics for a single cash flow, C = m * a * (D/D')^x
    where:
      C is the cashflow ($)
      m is a scalar multiplier
      a is the value of the widget, based on the D' volume sold
      D is the amount of widgets sold
      D' is the nominal amount of widgets sold
      x is the scaling factor
  """
  ##################
  # INITIALIZATION #
  ##################
  @classmethod
  def get_input_specs(cls):
    """
      Collects input specifications for this class.
      @ In, None
      @ Out, input_specs, InputData, specs
    """
    cf = InputData.parameterInputFactory('CashFlow')
    cf.description = r"""node for defining a CashFlow for a particular Component. This HERON
               CashFlow will be used to generate a TEAL CashFlow from RAVEN's TEAL plugin. Note a CashFlow generally
               takes the form $C = \alpha \left(\frac{D}{D'}\right)^x$, aggregated depending
               on the \xmlAttr{type}. For more information, see the TEAL plugin for RAVEN."""

    cf.addParam('name', param_type=InputTypes.StringType, required=True,
        descr=r"""the name by which this CashFlow will be identified as part of this component. The
              general name is prefixed by the component name, such as ComponentName$\vert$CashFlowName. """)
    cf_type_enum = InputTypes.makeEnumType('CFType', 'CFType', ['one-time', 'repeating'])
    cf.addParam('type', param_type=cf_type_enum, required=True,
        descr=r"""the type of CashFlow to calculate. \xmlString(one-time) is suitable for capital
              expenditure CashFlows, while \xmlString(repeating) is used for repeating costs such as
              operations and maintenance (fixed or variable), market sales, or similar.""")
    cf.addParam('taxable', param_type=InputTypes.BoolType, required=True,
        descr=r"""determines whether this CashFlow is taxed every cycle. """)
    cf.addParam('inflation', param_type=InputTypes.StringType, required=True,
        descr=r"""determines how inflation affects this CashFlow every cycle. See the CashFlow submodule
              of RAVEN.""")
    cf.addParam('mult_target', param_type=InputTypes.BoolType, required=True,
        descr=r"""indicates whether this parameter should be a target of the multiplication factor
              for NPV matching analyses.""")
    period_enum = InputTypes.makeEnumType('period_opts', 'period_opts', ['hour', 'year'])
    cf.addParam('period', param_type=period_enum, required=False,
        descr=r"""for a \xmlNode{CashFlow} with \xmlAttr{type} \xmlString{repeating}, indicates whether
              the CashFlow repeats every time step (\xmlString{hour}) or every cycle (\xmlString{year})).
              Generally, CashFlows such as fixed operations and maintenance costs are per-cycle, whereas
              variable costs such as fuel and maintenance as well as sales are repeated every time step.""")

    descr = r"""indicates the main driver for this CashFlow, such as the number of units sold
            or the size of the constructed unit. Corresponds to $D$ in the CashFlow equation."""
    driver = ValuedParams.factory.make_input_specs('driver', descr=descr, kind='post-dispatch')
    cf.addSub(driver)

    descr = r"""indicates the cash value of the reference number of units sold.
            corresponds to $\alpha$ in the CashFlow equation. If \xmlNode{reference_driver}
            is 1, then this is the price-per-unit for the CashFlow."""
    reference_price = ValuedParams.factory.make_input_specs('reference_price', descr=descr, kind='post-dispatch')
    cf.addSub(reference_price)

    descr = r"""determines the number of units sold to which the \xmlNode{reference_price}
            refers. Corresponds to $\prime D$ in the CashFlow equation. """
    reference_driver = ValuedParams.factory.make_input_specs('reference_driver', descr=descr, kind='post-dispatch')
    cf.addSub(reference_driver)

    descr = r"""determines the scaling factor for this CashFlow. Corresponds to $x$ in the CashFlow
            equation. If $x$ is less than one, the per-unit price decreases as the units sold increases
            above the \xmlNode{reference_driver}, and vice versa."""
    x = ValuedParams.factory.make_input_specs('scaling_factor_x', descr=descr, kind='post-dispatch')
    cf.addSub(x)

    depreciate = InputData.parameterInputFactory('depreciate', contentType=InputTypes.IntegerType)
    depreciate.descr = r"""indicates the number of cycles over which this CashFlow should be depreciated.
                       Depreciation schemes are assumed to be MACRS and available cycles are listed
                       in the CashFlow submodule of RAVEN."""
    cf.addSub(depreciate)

    return cf

  def __init__(self, component):
    """
      Constructor
      @ In, component, CashFlowUser instance, cash flow user to which this cash flow belongs
      @ Out, None
    """
    # assert component is not None # TODO is this necessary? What if it's not a component-based cash flow?
    self._component = component # component instance to whom this cashflow belongs, if any
    # equation values
    self._driver = None       # ValuedParam "quantity produced", D
    self._alpha = None        # ValuedParam "price per produced", a
    self._reference = None    # ValuedParam "where price is accurate", D'
    self._scale = None        # ValuedParam "economy of scale", x
    # other params
    self.name = None          # base name of cash flow
    self._type = None         # needed? one-time, yearly, repeating
    self._taxable = None      # apply tax or not
    self._inflation = None    # apply inflation or not
    self._mult_target = None  # not clear
    self._depreciate = None
    self._period = None       # period for recurring cash flows
    # other members
    self._signals = set()     # variable values needed for this cash flow
    self._crossrefs = defaultdict(dict)

  def read_input(self, item):
    """
      Sets settings from input file
      @ In, item, InputData.ParameterInput, parsed specs from user
      @ Out, None
    """
    self.name = item.parameterValues['name']
    # handle type directly here momentarily
    self._taxable = item.parameterValues['taxable']
    self._inflation = item.parameterValues['inflation']
    self._mult_target = item.parameterValues['mult_target']
    self._type = item.parameterValues['type']
    self._period = item.parameterValues.get('period', 'hour')
    # the remainder of the entries are ValuedParams, so they'll be evaluated as-needed
    for sub in item.subparts:
      if sub.getName() == 'driver':
        self._set_valued_param('_driver', sub)
      elif sub.getName() == 'reference_price':
        self._set_valued_param('_alpha', sub)
      elif sub.getName() == 'reference_driver':
        self._set_valued_param('_reference', sub)
      elif sub.getName() == 'scaling_factor_x':
        self._set_valued_param('_scale', sub)
      elif sub.getName() == 'depreciate':
        self._depreciate = sub.value

      else:
        raise IOError('Unrecognized "CashFlow" node: "{}"'.format(sub.getName()))

    # driver is required!
    if self._driver is None:
      raise IOError('No <driver> node provided for CashFlow {}!'.format(self.name))
    if self._alpha is None:
      raise IOError('No <reference_price> node provided for CashFlow {}!'.format(self.name))

    # defaults
    var_names = ['_reference', '_scale']
    for name in var_names:
      if getattr(self, name) is None:
        # TODO raise a warning?
        self._set_fixed_param(name, 1)

  # Not none set it to default 1
  def get_period(self):
    """
      Getter for Recurring cashflow period type.
      @ In, None
      @ Out, period, str, 'hourly' or 'yearly'
    """
    return self._period

  def _set_fixed_param(self, name, value):
    """
      Fixes a ValuedParam to have a constant value
      @ In, name, str, name of member to store on "self"
      @ In, value, float, value to set for ValuedParam
      @ Out, None
    """
    vp = ValuedParamHandler(name)
    vp.set_const_VP(value)
    setattr(self, name, vp)

  def _set_valued_param(self, name, spec):
    """
      Utilitly method to set ValuedParam members via reading input specifications.
      @ In, name, str, member variable name (e.g. self.<name>)
      @ In, spec, InputData params, input parameters
      @ Out, None
    """
    vp = ValuedParamHandler(name)
    signal = vp.read('CashFlow \'{}\''.format(self.name), spec, None) # TODO what "mode" to use?
    self._signals.update(signal)
    self._crossrefs[name] = vp
    # standard alias: redirect "capacity" variable
    if isinstance(vp, ValuedParams.factory.returnClass('variable')) and vp._raven_var == 'capacity':
      vp = self._component.get_capacity_param()
    setattr(self, name, vp)

  def get_alpha_extension(self):
    """
      creates multiplier for the valued shape the alpha cashflow parameter should be in
      @ In, None,
      @ Out, ext, multiplier for "alpha" values based on CashFlow type
    """
    life = self._component.get_economics().get_lifetime()
    if self._type == 'one-time':
      ext = np.zeros(life+1, dtype=float)
      ext[0] = 1.0
    elif self._type == 'repeating':
      ext = np.ones(life+1, dtype=float)
      ext[0] = 0.0
    else:
      raise NotImplementedError('type is: {}'.format(self._type))
    return ext

  def get_crossrefs(self):
    """
      Accessor for cross-referenced entities needed by this cashflow.
      @ In, None
      @ Out, crossrefs, dict, cross-referenced requirements dictionary
    """
    return self._crossrefs

  def set_crossrefs(self, refs):
    """
      Setter for cross-referenced entities needed by this cashflow.
      @ In, refs, dict, cross referenced entities
      @ Out, None
    """
    # set up pointers
    for attr, obj in refs.items():
      valued_param = self._crossrefs[attr]
      valued_param.set_object(obj)
    # check on VP setup
    for attr, vp in self._crossrefs.items():
      vp.crosscheck(self._component.get_interaction())

  def evaluate_cost(self, activity, values_dict):
    """
      Evaluates cost of a particular scenario provided by "activity".
      @ In, activity, pandas.Series, multi-indexed array of scenario activities
      @ In, values_dict, dict, additional values that may be needed to evaluate cost
      @ In, t, int, time index at which cost should be evaluated
      @ Out, cost, float, cost of activity
    """
    # note this method gets called a LOT, so speedups here are quite effective
    # add the activity to the dictionary
    values_dict['HERON']['activity'] = activity
    params = self.calculate_params(values_dict)
    return params['cost']

  def calculate_params(self, values_dict):
    """
      Calculates the value of the cash flow parameters.
      @ In, values_dict, dict, mapping from simulation variable names to their values (as floats or numpy arrays)
      @ Out, params, dict, dictionary of parameters mapped to values including the cost
    """
    # TODO maybe don't cast these as floats, as they could be symbolic expressions (seems unlikely)
    Dp = float(self._reference.evaluate(values_dict, target_var='reference_driver')[0]['reference_driver'])
    x = float(self._scale.evaluate(values_dict, target_var='scaling_factor_x')[0]['scaling_factor_x'])
    a = self._alpha.evaluate(values_dict, target_var='reference_price')[0]['reference_price']
    D = self._driver.evaluate(values_dict, target_var='driver')[0]['driver']
    cost = a * (D / Dp) ** x
    params = {'alpha': a, 'driver': D, 'ref_driver': Dp, 'scaling': x, 'cost': cost} # TODO float(cost) except in pyomo it's not a float
    return params

  def get_cashflow_params(self, values_dict, aliases, dispatches, years):
    """
      creates a param dict for initializing a CashFlows.CashFlow
      FIXME deprecated
      @ In, values_dict, dict, parameters dictionary
      @ In, aliases, dict, aliased names (unused)
      @ In, dispatches, dict, component activity
      @ In, years, int, years to obtain values for
      @ Out, params, dict, params needed for CashFlow
    """
    # OLD
    params = {'name': self.name,
              'reference': vals_dict['ref_driver'],
              'driver': vals_dict['driver'],
              'alpha': vals_dict['alpha'],
              'X': vals_dict['scaling'],
              'mult_target': self._mult_target,
              'inflation': self._inflation,
              'multiply': None,
              'tax': self._taxable,
              }
    return params




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Holds the template information for creating LCOE SWEEP OPT input files.
"""
import os
import sys
import copy
import shutil
import time
import xml.etree.ElementTree as ET
import itertools as it

import numpy as np
import dill as pk

# load utils
sys.path.append(os.path.join(os.path.dirname(__file__), '..', '..'))
from HERON.src.base import Base
import HERON.src._utils as hutils
sys.path.pop()

# get raven location
RAVEN_LOC = os.path.abspath(os.path.join(hutils.get_raven_loc(), "ravenframework"))
try:
  import TEAL.src
except ModuleNotFoundError:
  CF_LOC = hutils.get_cashflow_loc(raven_path=RAVEN_LOC)
  if CF_LOC is None:
    raise RuntimeError('TEAL has not been found!\n' +
                       f'Check TEAL installation for the RAVEN at "{RAVEN_LOC}"')

  sys.path.append(os.path.join(CF_LOC, '..'))
from TEAL.src.main import getProjectLength
from TEAL.src import CashFlows

sys.path.append(os.path.join(RAVEN_LOC, '..'))
from ravenframework.utils import xmlUtils
from ravenframework.InputTemplates.TemplateBaseClass import Template as TemplateBase
sys.path.pop()

class Template(TemplateBase, Base):
  """
    Template for lcoe sweep opt class
    This templates the workflow split into sweeping over unit capacities
    in an OUTER run while optimizing unit dispatch in a INNER run.

    As designed, the ARMA stochastic noise happens entirely on the INNER,
    for easier parallelization.
  """

  # dynamic naming templates
  TemplateBase.addNamingTemplates({'jobname'        : '{case}_{io}',
                                   'stepname'       : '{action}_{subject}',
                                   'variable'       : '{unit}_{feature}',
                                   'dispatch'       : 'Dispatch__{component}__{tracker}__{resource}',
                                   'data object'    : '{source}_{contents}',
                                   'distribution'   : '{unit}_{feature}_dist',
                                   'ARMA sampler'   : '{rom}_sampler',
                                   'lib file'       : 'heron.lib', # TODO use case name?
                                   'cashfname'      : '_{component}{cashname}',
                                   're_cash'        : '_rec_{period}_{driverType}{driverName}',
                                   'cluster_index'  : '_ROM_Cluster',
                                  })

  # template nodes
  dist_template = xmlUtils.newNode('Uniform')
  dist_template.append(xmlUtils.newNode('lowerBound'))
  dist_template.append(xmlUtils.newNode('upperBound'))

  var_template = xmlUtils.newNode('variable')
  var_template.append(xmlUtils.newNode('distribution'))

  ############
  # API      #
  ############
  def __repr__(self):
    """
      String representation of this Handler and its VP
      @ In, None
      @ Out, repr, str, string representation
    """
    msg = f'<Template Driver>'
    return msg

  def __init__(self, **kwargs):
    """
      Constructor
      @ In, None
      @ Out, None
    """
    Base.__init__(self, **kwargs)
    here = os.path.dirname(os.path.abspath(sys.modules[self.__class__.__module__].__file__))
    self._template_path = here
    self._template_inner_path = None
    self._template_outer_path = None
    self._template_cash_path = None
    self._template_cash = None
    self._template_inner = None
    self._template_outer = None
    self.__case = None
    self.__components = None
    self.__sources = None
    self.__sweep_vars = []

  def loadTemplate(self, path):
    """
      Loads RAVEN template files from source.
      @ In, path, str, relative path to templates
      @ Out, None
    """
    rel_path = os.path.join(self._template_path, path)
    self._template_inner_path = os.path.join(rel_path, 'inner.xml')
    self._template_outer_path = os.path.join(rel_path, 'outer.xml')
    self._template_cash_path = os.path.join(rel_path, 'cash.xml')

    self._template_inner, _ = xmlUtils.loadToTree(self._template_inner_path, preserveComments=True)
    self._template_outer, _ = xmlUtils.loadToTree(self._template_outer_path, preserveComments=True)
    self._template_cash, _ = xmlUtils.loadToTree(self._template_cash_path, preserveComments=True)

  def createWorkflow(self, case, components, sources):
    """
      Create workflow XMLs
      @ In, case, HERON case, case instance for this sim
      @ In, components, list, HERON component instances for this sim
      @ In, sources, list, HERON source instances for this sim
      @ Out, inner, XML Element, root node for inner
      @ Out, outer, XML Element, root node for outer
      @ Out, cash, XML Element, root node for cashflow input
    """
    # store pieces
    self.__case = case
    self.__components = components
    self.__sources = sources
    # initialize case economics
    case.load_econ(components)
    # load a copy of the template
    inner = copy.deepcopy(self._template_inner)
    outer = copy.deepcopy(self._template_outer)
    cash = copy.deepcopy(self._template_cash)
    # modify the templates
    inner = self._modify_inner(inner, case, components, sources)
    outer = self._modify_outer(outer, case, components, sources)
    cash = self._modify_cash(cash, case, components, sources)
    return inner, outer, cash

  def writeWorkflow(self, templates, destination, run=False):
    """
      Write outer and inner RAVEN workflows.
      @ In, templates, list, modified XML roots
      @ In, destination, str, path to write workflows to
      @ In, run, bool, if True then attempt to run the workflows
      @ Out, None
    """
    # TODO use destination?
    # write templates
    inner, outer, cash = templates
    outer_file = os.path.abspath(os.path.join(destination, 'outer.xml'))
    inner_file = os.path.abspath(os.path.join(destination, 'inner.xml'))
    cash_file = os.path.abspath(os.path.join(destination, 'cash.xml'))
    self.raiseAMessage('========================')
    self.raiseAMessage('HERON: writing files ...')
    self.raiseAMessage('========================')
    msg_format = 'Wrote "{1}" to "{0}/"'
    with open(outer_file, 'w') as f:
      f.write(xmlUtils.prettify(outer))
    self.raiseAMessage(msg_format.format(*os.path.split(outer_file)))
    with open(inner_file, 'w') as f:
      f.write(xmlUtils.prettify(inner))
    self.raiseAMessage(msg_format.format(*os.path.split(inner_file)))
    with open(cash_file, 'w') as f:
      f.write(xmlUtils.prettify(cash))
    self.raiseAMessage(msg_format.format(*os.path.split(cash_file)))
    # write library of info so it can be read in dispatch during inner run
    data = (self.__case, self.__components, self.__sources)
    lib_file = os.path.abspath(os.path.join(destination, self.namingTemplates['lib file']))
    with open(lib_file, 'wb') as lib:
      pk.dump(data, lib)
    self.raiseAMessage(msg_format.format(*os.path.split(lib_file)))
    # copy "write_inner.py", which has the denoising and capacity fixing algorithms
    conv_src = os.path.abspath(os.path.join(self._template_path, 'write_inner.py'))
    conv_file = os.path.abspath(os.path.join(destination, 'write_inner.py'))
    shutil.copyfile(conv_src, conv_file)
    self.raiseAMessage(msg_format.format(*os.path.split(conv_file)))
    # run, if requested
    if run:
      self.runWorkflow(destination)

  ############
  # UTILS    #
  ############
  ##### OUTER #####
  # Right now we modify outer by RAVEN structure, rather than HERON features
  def _modify_outer(self, template, case, components, sources):
    """
      Defines modifications to the outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, template, xml.etree.ElementTree.Element, modified template
    """
    template.set('verbosity', case.get_verbosity())
    self._modify_outer_mode(template, case, components, sources)
    self._modify_outer_runinfo(template, case)
    self._modify_outer_vargroups(template, case, components, sources)
    self._modify_outer_databases(template, case)
    self._modify_outer_dataobjects(template, case, components)
    self._modify_outer_files(template, case, sources)
    self._modify_outer_models(template, case, components, sources)
    self._modify_outer_outstreams(template, case, components, sources)
    self._modify_outer_samplers(template, case, components)
    self._modify_outer_optimizers(template, case)
    self._modify_outer_steps(template, case, components, sources)
    return template

  def _modify_outer_mode(self, template, case, components, sources):
    """
      Defines major (entity-level) modifications to outer.xml RAVEN input file due to case mode
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    if case.get_mode() == 'sweep' or case.debug['enabled']:
      template.remove(template.find('Optimizers'))
    elif case.get_mode() == 'opt':
      template.remove(template.find('Samplers'))

  def _modify_outer_runinfo(self, template, case):
    """
      Defines modifications to the RunInfo of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    run_info = template.find('RunInfo')
    case_name = self.namingTemplates['jobname'].format(case=case.name, io='o')
    run_info.find('JobName').text = case_name
    run_info.find('WorkingDir').text = case_name
    if case.debug['enabled']:
      seq = run_info.find('Sequence')
      seq.text = 'debug'
      self._updateCommaSeperatedList(seq, 'debug_output')
    elif case.get_mode() == 'sweep':
      run_info.find('Sequence').text = 'sweep'
    elif case.get_mode() == 'opt':
      run_info.find('Sequence').text = 'optimize, plot'
    # parallel
    if case.outerParallel:
      # set outer batchsize and InternalParallel
      batchSize = run_info.find('batchSize')
      batchSize.text = f'{case.outerParallel}'
      run_info.append(xmlUtils.newNode('internalParallel', text='True'))
    if case.useParallel:
      #XXX this doesn't handle non-mpi modes like torque or other custom ones
      mode = xmlUtils.newNode('mode', text='mpi')
      mode.append(xmlUtils.newNode('runQSUB'))
      if 'memory' in case.parallelRunInfo:
        mode.append(xmlUtils.newNode('memory', text=case.parallelRunInfo.pop('memory')))
      for sub in case.parallelRunInfo:
        run_info.append(xmlUtils.newNode(sub, text=str(case.parallelRunInfo[sub])))
      run_info.append(mode)
    if case.innerParallel:
      run_info.append(xmlUtils.newNode('NumMPI', text=case.innerParallel))

  def _modify_outer_vargroups(self, template, case, components, sources):
    """
      Defines modifications to the VariableGroups of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    var_groups = template.find('VariableGroups')
    # capacities
    caps = var_groups[0]
    var_list = []

    # Add component opt vars
    for comp in components:
      comp_cap_type = comp.get_capacity(None, raw=True).type
      if comp_cap_type  not in ['Function', 'ARMA', 'SyntheticHistory']:
        var_list.append(f'{comp.name}_capacity')

    # Add dispatch opt vars
    for var in case.dispatch_vars.keys():
      var_list.append(f'{var}_dispatch')
    caps.text = ', '.join(var_list)

    # outer results
    group_outer_results = var_groups.find(".//Group[@name='GRO_outer_results']")
    # add required defaults
    default_stats = [f'mean_{case._metric}', f'std_{case._metric}', f'med_{case._metric}']
    for stat in default_stats:
      self._updateCommaSeperatedList(group_outer_results, stat)
    # make sure user provided statistics beyond defaults get there
    if any(stat not in ['expectedValue', 'sigma', 'median'] for stat in case._result_statistics):
      stats_list = self._build_result_statistic_names(case)
      for stat_name in stats_list:
        if stat_name not in default_stats:
          self._updateCommaSeperatedList(group_outer_results, stat_name)
    # sweep mode has default variable names
    elif case.get_mode() == 'sweep':
      sweep_default = [f'mean_{case._metric}', f'std_{case._metric}', f'med_{case._metric}', f'max_{case._metric}',
                       f'min_{case._metric}', f'perc_5_{case._metric}', f'perc_95_{case._metric}',
                       f'samp_{case._metric}', f'var_{case._metric}']
      for sweep_name in sweep_default:
        if sweep_name not in default_stats:
          self._updateCommaSeperatedList(group_outer_results, sweep_name)
    # opt mode adds optimization variable if not already there
    if (case.get_mode() == 'opt') and (case._optimization_settings is not None):
      new_metric_outer_results = self._build_opt_metric_out_name(case)
      if (new_metric_outer_results != 'missing') and (new_metric_outer_results not in group_outer_results.text):
        # additional results statistics have been requested, add this metric if not already present
        self._updateCommaSeperatedList(group_outer_results, new_metric_outer_results, position=0)

    # labels group
    if case.get_labels():
      case_labels = ET.SubElement(var_groups, 'Group', attrib={'name': 'GRO_case_labels'})
      case_labels.text = ', '.join([f'{key}_label' for key in case.get_labels().keys()])
    if case.debug['enabled']:
      # expected dispatch, ARMA outputs
      # -> dispatch results
      group = var_groups.find(".//Group[@name='GRO_outer_debug_dispatch']")
      for component in components:
        name = component.name
        for tracker in component.get_tracking_vars():
          for resource in component.get_resources():
            var_name = self.namingTemplates['dispatch'].format(component=name, tracker=tracker, resource=resource)
            self._updateCommaSeperatedList(group, var_name)
      # -> synthetic histories?
      group = var_groups.find(".//Group[@name='GRO_outer_debug_synthetics']")
      for source in sources:
        if source.is_type('ARMA'):
          synths = source.get_variable()
          for synth in synths:
            if not group.text or synth not in group.text.split(','):
              self._updateCommaSeperatedList(group, synth)

  def _modify_outer_databases(self, template, case):
    """
      Defines modifications to the Databases of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    if case.debug['enabled']:
      DBs = xmlUtils.newNode('Databases')
      template.append(DBs)
      attrs = {'name': 'dispatch', 'readMode': 'overwrite', 'directory': ''}
      db = xmlUtils.newNode('NetCDF', attrib=attrs)
      db.append(xmlUtils.newNode('variables', text='GRO_outer_debug_dispath,GRO_outer_debug_synthetics'))
      DBs.append(db)

  def _modify_outer_dataobjects(self, template, case, components):
    """
      Defines modifications to the DataObjects of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ Out, None
    """
    DOs = template.find('DataObjects')
    # labels pass to inner
    if case.get_labels():
      for node in DOs:
        if node.get('name') == 'grid':
          input_node = node.find('Input')
          input_node.text += ', GRO_case_labels'
    # remove opt components if not used
    if case.get_mode() == 'sweep' or case.debug['enabled']:
      self._remove_by_name(DOs, ['opt_eval', 'opt_soln'])
    elif case.get_mode() == 'opt':
      self._remove_by_name(DOs, ['grid'])
    # debug mode
    if case.debug['enabled']:
      # add debug dispatch output dataset
      debug_gro = ['GRO_outer_debug_dispatch', 'GRO_outer_debug_synthetics']
      deps = {self.__case.get_time_name(): debug_gro,
              self.namingTemplates['cluster_index']: debug_gro,
              self.__case.get_year_name(): debug_gro}
      self._create_dataobject(DOs, 'DataSet', 'dispatch',
                              inputs=['scaling'],
                              outputs=debug_gro,
                              depends=deps)

  def _modify_outer_files(self, template, case, sources):
    """
      Defines modifications to the Files of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    files = template.find('Files')
    multiruns = template.find('Steps').findall('MultiRun')
    for run in multiruns:
      # This relies on the fact that the template is hardcoded so that
      # the MultiRun 'name' attribute is either 'optimize' or 'sweep'.
      if case.get_mode() in run.attrib['name']:
        step = run
    # modify path to inner
    inner = files.find('Input') # NOTE assuming it's the first file in the template
    inner.text = '../inner.xml'
    # add other files needed by inner (functions, armas, etc)
    for source in sources:
      if source.is_type('Function'):
        # add it to the list of things that have to be transferred
        files = template.find('Files')
        src = xmlUtils.newNode('Input', attrib={'name': source.name}, text='../'+source._source)
        files.append(src)
        # add it to the Step inputs so it gets carried along
        inp = self._assemblerNode('Input', 'Files', '', source.name)
        step.insert(0, inp)

  def _modify_outer_models(self, template, case, components, sources):
    """
      Defines modifications to the Models of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    raven = template.find('Models').find('Code')
    # executable
    raven_exec = raven.find('executable')
    raven_exec_guess = os.path.abspath(os.path.join(RAVEN_LOC, '..', 'raven_framework'))
    if os.path.exists(raven_exec_guess):
      raven_exec.text = raven_exec_guess
    elif shutil.which("raven_framework") is not None:
      raven_exec.text = "raven_framework"
    else:
      raise RuntimeError("raven_framework not in PATH and not at "+raven_exec_guess)
    # conversion script
    conv = raven.find('conversion').find('input')
    conv.attrib['source'] = '../write_inner.py'

    # NOTE: if we find any CSVs in sources, we know the structure of our inner
    # has changed quite a bit. This is because we abandon the EnsembleModel &
    # MonteCarlo samplers in favor of a ExternalModel & CustomSampler when
    # using Static Histories instead of a Synthetic History. Also we do not
    # anticpate/allow users to mix the use of Static & Synthetic Histories.
    if any(x.is_type("CSV") for x in sources):
      # NOTE: this chunk of code does the initial footwork for switching
      # to Static Histories. The rest of the required changes are completed
      # in the _modify_inner_static_history() method.
      text = 'Samplers|CustomSampler@name:mc_arma_dispatch|constant@name:{}'
      # Remove anything having to do with 'denoises', it's no longer needed.
      raven.remove(raven.find(".//alias[@variable='denoises']"))
      denoises_parent = template.find(".//constant[@name='denoises']/..")
      denoises_parent.remove(denoises_parent.find(".//constant[@name='denoises']"))
      # Remove any GRO_final_return vars that compute Sigma or Var (e.g. var_NPV)
      final_return_vars = template.find('.//VariableGroups/Group[@name="GRO_outer_results"]')
      new_final_return_vars = [var for var in final_return_vars.text.split(", ") if "std" not in var and "var" not in var]
      final_return_vars.text = ', '.join(new_final_return_vars)
    else:
      text = 'Samplers|MonteCarlo@name:mc_arma_dispatch|constant@name:{}'

    for component in components:
      name = component.name
      attribs = {'variable': f'{name}_capacity', 'type':'input'}
      new = xmlUtils.newNode('alias', text=text.format(name + '_capacity'), attrib=attribs)
      raven.append(new)

    # Now we check for any non-component dispatch variables and assign aliases
    for name in case.dispatch_vars.keys():
      attribs = {'variable': f'{name}_dispatch', 'type':'input'}
      new = xmlUtils.newNode('alias', text=text.format(name + '_dispatch'), attrib=attribs)
      raven.append(new)

    # label aliases placed inside models
    for label in case.get_labels():
      attribs = {'variable': f'{label}_label', 'type':'input'}
      new = xmlUtils.newNode('alias', text=text.format(label + '_label'), attrib=attribs)
      raven.append(new)

    # if debug, grab the dispatch output instead of the summary
    if case.debug['enabled']:
      raven.find('outputDatabase').text = 'disp_full'

    # data handling: inner to outer data format
    if case.data_handling['inner_to_outer'] == 'csv':
      # swap the outputDatabase to outputExportOutStreams
      output_node = template.find('Models').find('Code').find('outputDatabase')
      output_node.tag = 'outputExportOutStreams'
      # no need to change name, as database and outstream have the same name


  def _modify_outer_outstreams(self, template, case, components, sources):
    """
      Defines modifications to the OutStreams of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    OSs = template.find('OutStreams')
    # remove opt if not used
    if case.get_mode() == 'sweep' or case.debug['enabled']:
      self._remove_by_name(OSs, ['opt_soln'])
    elif case.get_mode() == 'opt':
      self._remove_by_name(OSs, ['sweep'])
      # update plot 'opt_path' if necessary
      new_opt_objective = self._build_opt_metric_out_name(case)
      opt_path_plot_vars = OSs.find(".//Plot[@name='opt_path']").find('vars')
      if (new_opt_objective != 'missing') and (new_opt_objective not in opt_path_plot_vars.text):
        opt_path_plot_vars.text = opt_path_plot_vars.text.replace(f'mean_{case._metric}', new_opt_objective)
    # debug mode
    if case.debug['enabled']:
      # modify normal metric output
      out = OSs.findall('Print')[0]
      out.attrib['name'] = 'dispatch_print'
      out.find('source').text = 'dispatch'
      # handle dispatch plots for debug mode
      if case.debug['dispatch_plot']:
        out_plot = ET.SubElement(OSs, 'Plot', attrib={'name': 'dispatchPlot', 'subType': 'HERON.DispatchPlot'})
        out_plot_source = ET.SubElement(out_plot, 'source')
        out_plot_source.text = 'dispatch'
        out_plot_macro = ET.SubElement(out_plot, 'macro_variable')
        out_plot_macro.text = case.get_year_name()
        out_plot_micro = ET.SubElement(out_plot, 'micro_variable')
        out_plot_micro.text = case.get_time_name()
        out_plot_signals = ET.SubElement(out_plot, 'signals')
        signals = set()
        for source in sources:
          new = source.get_variable()
          if new is not None:
            signals.update(set(new))
        out_plot_signals.text = ', '.join(signals)

  def _modify_outer_samplers(self, template, case, components):
    """
      Defines modifications to the Samplers/Optimizers of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ Out, None
    """
    dists_node = template.find('Distributions')
    if case.get_mode() == 'sweep' or case.debug['enabled']:
      samps_node = template.find('Samplers/Grid')
    else:
      samps_node = template.find('Optimizers/GradientDescent')
    if case.debug['enabled']:
      samps_node.tag = 'MonteCarlo'
      samps_node.attrib['name'] = 'mc'
      init = xmlUtils.newNode('samplerInit')
      init.append(xmlUtils.newNode('limit', text=1))
      samps_node.append(init)

    # NOTE: There is a chance we removed the denoises variable earlier.
    # If it was removed, that means we are using a StaticHistory.
    if samps_node.find('.//constant[@name="denoises"]') is not None:
      samps_node.find('.//constant[@name="denoises"]').text = str(case.get_num_samples())
    # add sweep variables to input

    ## TODO: Refactor this portion with the below portion to handle
    ## all general cases instead of only two.
    for key, value in case.get_labels().items():
      var_name = self.namingTemplates['variable'].format(unit=key, feature='label')
      samps_node.append(xmlUtils.newNode('constant', text=value, attrib={'name': var_name}))

    if case.debug['enabled']:
      sampler = 'mc'
    elif case.get_mode() == 'sweep':
      sampler = 'grid'
    else:
      sampler = 'opt'

    for key, value in case.dispatch_vars.items():
      var_name = self.namingTemplates['variable'].format(unit=key, feature='dispatch')
      vals = value.get_value(debug=case.debug['enabled'])
      if isinstance(vals, list):
        dist, xml = self._create_new_sweep_capacity(key, var_name, vals, sampler)
        dists_node.append(dist)
        if case.get_mode() == 'sweep':
          samps_node.append(xml)
        else:
          samps_node.append(xml)

    for component in components:
      interaction = component.get_interaction()
      # NOTE this algorithm does not check for everthing to be swept! Future work could expand it.
      # This is approached by the labels feature above
      ## Currently checked: Component.Interaction.Capacity
      ## --> this really needs to be made generic for all kinds of valued params!
      name = component.name
      var_name = self.namingTemplates['variable'].format(unit=name, feature='capacity')
      cap = interaction.get_capacity(None, raw=True)
      # do we already know the capacity values?
      if cap.is_parametric():
        vals = cap.get_value(debug=case.debug['enabled'])
        # is the capacity variable being swept over?
        if isinstance(vals, list):
          # make new Distribution, Sampler.Grid.variable
          dist, xml = self._create_new_sweep_capacity(name, var_name, vals, sampler)
          dists_node.append(dist)
          samps_node.append(xml)
          # NOTE assumption (input checked): only one interaction per component
        # if not being swept, then it's just a fixed value.
        else:
          samps_node.append(xmlUtils.newNode('constant', text=vals, attrib={'name': var_name}))
      else:
        # this capacity will be evaluated by ARMA/Function, and doesn't need to be added here.
        pass
    if case.outerParallel == 0 and case.useParallel:
      #XXX if we had a way to calculate this ahead of time,
      # this could be done in _modify_outer_runinfo
      #Need to update the outerParallel number
      run_info = template.find('RunInfo')
      case.outerParallel = len(self.__sweep_vars) + 1
      #XXX duplicate of code in _modify_outer_runinfo
      batchSize = run_info.find('batchSize')
      batchSize.text = f'{case.outerParallel}'
      run_info.append(xmlUtils.newNode('internalParallel', text='True'))

  def _modify_outer_optimizers(self, template, case):
    """
      Defines modifications to the Optimizers of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """

    # only modify if optimization_settings is in Case
    if (case.get_mode() == 'opt') and (case._optimization_settings is not None) and (not case.debug['enabled']):  # TODO there should be a better way to handle the debug case
      # TODO will the optimizer always be GradientDescent?
      opt_node = template.find('Optimizers').find(".//GradientDescent[@name='cap_opt']")
      new_opt_objective = self._build_opt_metric_out_name(case)
      # swap out objective if necessary
      opt_node_objective = opt_node.find('objective')
      if (new_opt_objective != 'missing') and (new_opt_objective != opt_node_objective.text):
        opt_node_objective.text = new_opt_objective
      # swap out samplerInit values (only type implemented now)
      sampler_init = opt_node.find('samplerInit')
      type_node = sampler_init.find('type')
      try:
        type_node.text = case._optimization_settings['type']
      except KeyError:
        # type was not provided, so use the default value
        metric_raven_name = case._optimization_settings['metric']['name']
        type_node.text = case.metrics_mapping[metric_raven_name]['optimization_default']

      # swap out convergence values (only persistence implemented now)
      convergence = opt_node.find('convergence')
      persistence_node = convergence.find('persistence')
      try:
        persistence_node.text = str(case._optimization_settings['persistence'])
      except KeyError:
        # persistence was not provided, so use the default value
        pass

      # update convergence criteria, adding nodes as necessary
      convergence_settings = case._optimization_settings.get('convergence', {})
      for k, v in convergence_settings.items():
        node = convergence.find(k)  # will return None if subnode is not found
        if node is None:
          convergence.append(ET.Element(k))
          node = convergence.find(k)
        node.text = v

  def _modify_outer_steps(self, template, case, components, sources):
    """
      Defines modifications to the Steps of outer.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    steps = template.find('Steps')
    # clear out optimization if not used
    if case.get_mode() == 'sweep' or case.debug['enabled']:
      self._remove_by_name(steps, ['optimize', 'plot'])
    elif case.get_mode() == 'opt':
      self._remove_by_name(steps, ['sweep'])
    if case.debug['enabled']:
      # repurpose the sweep multirun
      sweep = steps.findall('MultiRun')[0]
      sweep.attrib['name'] = 'debug'
      sweep.find('Sampler').attrib['type'] = 'MonteCarlo'
      sweep.find('Sampler').text = 'mc'
      # remove the BasicStats collector and printer
      to_remove = []
      for output in sweep.findall('Output'):
        if output.text in ['grid', 'sweep']:
          to_remove.append(output)
      for node in to_remove:
        sweep.remove(node)
      # add debug dispatch collector and printer
      sweep.append(self._assemblerNode('Output', 'DataObjects', 'DataSet', 'dispatch'))
      sweep.append(self._assemblerNode('Output', 'Databases', 'NetCDF', 'dispatch'))
      # add an output step to print/plot summaries
      io_step = ET.SubElement(steps, 'IOStep', attrib={'name': 'debug_output'})
      io_input = ET.SubElement(io_step, 'Input', attrib={'class': 'DataObjects', 'type': 'DataSet'})
      io_input.text = 'dispatch'
      io_step.append(self._assemblerNode('Output', 'OutStreams', 'Print', 'dispatch_print'))
      if case.debug['dispatch_plot']:
        io_output = ET.SubElement(io_step, 'Output', attrib={'class': 'OutStreams', 'type': 'Plot'})
        io_output.text = 'dispatchPlot'

  def _create_new_sweep_capacity(self, comp_name, var_name, capacities, sampler):
    """
      for OUTER, creates new distribution and variable for grid/opt sampling
      @ In, comp_name, str, name of component
      @ In, var_name, str, name of capacity variable
      @ In, capacities, list, float list of capacities to sweep/opt over
      @ In, sampler, string, which sampler to assume (grid, mc, opt)
      @ Out, dist, xml.etree.ElementTree,Element, XML for distribution
      @ Out, xml, xml.etree.ElementTree,Element, XML for sampled variable
    """
    # distribution
    if 'capacity' in var_name:
      dist_name = self.namingTemplates['distribution'].format(unit=comp_name, feature='capacity')
    else:
      dist_name = self.namingTemplates['distribution'].format(unit=comp_name, feature='dispatch')
    dist = copy.deepcopy(self.dist_template)
    dist.attrib['name'] = dist_name
    min_cap = min(capacities)
    max_cap = max(capacities)
    dist.find('lowerBound').text = str(min_cap)
    dist.find('upperBound').text = str(max_cap)
    xml = copy.deepcopy(self.var_template)
    xml.attrib['name'] = var_name
    self.__sweep_vars.append(var_name)
    xml.find('distribution').text = dist_name
    if sampler == 'grid':
      caps = ' '.join(str(x) for x in sorted(capacities))
      xml.append(xmlUtils.newNode('grid', attrib={'type':'value', 'construction':'custom'}, text=caps))
    elif sampler == 'opt':
      # initial value
      delta = max_cap - min_cap
      # start at 5% away from 0
      if max_cap > 0:
        initial = min_cap + 0.05 * delta
      else:
        initial = max_cap - 0.05 * delta
      xml.append(xmlUtils.newNode('initial', text=initial))
    return dist, xml



  ##### INNER #####
  # Right now we modify inner by HERON features, rather than RAVEN structure
  def _modify_inner(self, template, case, components, sources):
    """
      Defines modifications to the inner.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, template, xml.etree.ElementTree.Element, modified template
    """
    input_filepath = os.path.abspath((os.path.dirname(__file__)))
    input_filepath = input_filepath+'/../src/DispatchManager'
    ext_model = template.find('Models').find('ExternalModel')
    #ModuleToLoad not needed for HERON.DispatchManager plugin
    if 'ModuleToLoad' in ext_model.attrib:
      ext_model.attrib.pop('ModuleToLoad')
    ext_model.set('subType','HERON.DispatchManager')
    self._modify_inner_runinfo(template, case)
    self._modify_inner_sources(template, case, components, sources)
    # NOTE: this HAS to come before modify_inner_denoisings,
    #       because we'll be copy-pasting these for each denoising --> or wait, maybe that's for the Outer to do!
    self._modify_inner_components(template, case, components)
    self._modify_inner_caselabels(template, case)
    self._modify_inner_time_vars(template, case)
    self._modify_inner_result_statistics(template, case)
    self._modify_inner_optimization_settings(template, case)
    self._modify_inner_data_handling(template, case)
    if case.debug['enabled']:
      self._modify_inner_debug(template, case, components)
    self._modify_inner_static_history(template, case, sources)
    # TODO modify based on resources ... should only need if units produce multiple things, right?
    # TODO modify CashFlow input ... this will be a big undertaking with changes to the inner.
    ## Maybe let the user change them? but then we don't control the variable names. We probably have to do it.
    return template

  def _modify_inner_static_history(self, template, case, sources):
    """
      Modify entire Inner template if using StaticHistory.

      Using a static history changes many different aspects of the outer and
      inner templates. This function assumes that it will only find ONE csv in
      sources, otherwise it will make the modifications twice and could cause errors.

      @ In, template, ET.Element, root of XML template to modify
      @ In, case, HERON.Case, case object of current simulation
      @ In, sources, List[HERON.Placeholders], data generator sources for current simulation.
      @ Out, None
    """
    # If no CSV found, this won't run
    for source in filter(lambda x: x.is_type("CSV"), sources):
      # Add CSV file reference to <Files>
      csv_file = xmlUtils.newNode('Input', attrib={'name': source.name}, text=source._target_file)
      template.find("Files").append(csv_file)

      # Update <Sequence> to read the csv into memory first
      self._updateCommaSeperatedList(template.find(".//RunInfo/Sequence"), "read_static", position=0)

      # Create a new <IOStep> that is the instructions <Sequence> will run
      new_step = xmlUtils.newNode('IOStep', attrib={'name': 'read_static'})
      new_step.append(self._assemblerNode('Input', 'Files', '', source.name))
      new_step.append(self._assemblerNode('Output', 'DataObjects', 'DataSet', 'input'))
      template.find('Steps').append(new_step)

      # Change ExternalModel type in MultiRun Steps
      multi_run = template.find('.//Steps/MultiRun[@name="arma_sampling"]')
      multi_run.find("Sampler").attrib["type"] = "CustomSampler"
      multi_run.find('.//Model[@type="EnsembleModel"]').text = "dispatch"
      multi_run.find('.//Model[@type="EnsembleModel"]').attrib['type'] = "ExternalModel"

      # Modify <Group> node containing PP statistics. Remove all STD and VAR variables.
      gro_final_return = template.find('.//VariableGroups/Group[@name="GRO_final_return"]')
      new_return_vars = [var for var in gro_final_return.text.split(", ") if "std" not in var and "var" not in var]
      gro_final_return.text = ', '.join(new_return_vars)

      # Create a new <DataObject> that will store the csv data
      data_objs = template.find("DataObjects")
      new_data_set = xmlUtils.newNode("DataSet", attrib={"name": "input"})
      new_data_set.append(xmlUtils.newNode("Input", text=', '.join([case.get_time_name(), case.get_year_name()])))
      new_data_set.append(xmlUtils.newNode("Output", text=', '.join(source.get_variable())))
      for var in [case.get_year_name(), case.get_time_name()]:
        new_data_set.append(xmlUtils.newNode("Index", attrib={"var": var}, text=', '.join(source.get_variable())))
      data_objs.append(new_data_set)

      # Modify <Models> by removing EnsembleModel and changing ExternalModel
      models = template.find("Models")
      for var in source.get_variable():
        self._updateCommaSeperatedList(models.find('.//ExternalModel[@name="dispatch"]/variables'), var)

      self.raiseAMessage("Using Static History - replacing EnsembleModel with CustomSampler strategy")
      models.remove(models.find('.//EnsembleModel[@name="sample_and_dispatch"]'))

      # Remove PP Statistics that are no longer needed
      self.raiseAMessage(f'Using Static History - removing unneeded post-processor statistics "sigma" & "variance"')
      post_proc = models.find(".//PostProcessor")
      for sigma_node in it.chain(post_proc.findall(".//sigma"), post_proc.findall(".//variance")):
        post_proc.remove(sigma_node)

      # Modify <Samplers> to get rid of MonteCarlo reference in favor of CustomSampler
      samps = template.find("Samplers")
      monte_carlo = samps.find('.//MonteCarlo[@name="mc_arma_dispatch"]')
      monte_carlo.insert(0, self._assemblerNode("Source", "DataObjects", "DataSet", "input"))
      for var in it.chain([case.get_year_name(), case.get_time_name()], source.get_variable()):
        # We make the bold assumption here that our CSV contains a "Year" and "Time" variable.
        var_node = xmlUtils.newNode("variable", attrib={"name": var})
        monte_carlo.append(var_node)
      monte_carlo.remove(monte_carlo.find(".//samplerInit"))
      monte_carlo.tag = "CustomSampler"

  def _modify_inner_caselabels(self, template, case):
    """
      Create GRO_case_labels VariableGroup if labels have been provided.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    if case.get_labels():
      var_groups = template.find('VariableGroups')
      case_labels = ET.SubElement(var_groups, 'Group', attrib={'name': 'GRO_case_labels'})
      case_labels.text = ', '.join([f'{key}_label' for key in case.get_labels().keys()])
      ## Since <label> is optional, we don't want to hard-code it into
      ## the template files. So we will create it as needed and then
      ## modify GRO_armasamples_in_scalar to contain the group.
      for node in var_groups:
        if node.get('name') in ['GRO_armasamples_in_scalar', 'GRO_dispatch_in_scalar']:
          node.text += ', GRO_case_labels'
      # Add case labels to Sampler node as well.
      mc = template.find('Samplers').find('MonteCarlo')
      for key, value in case.get_labels().items():
        label_name = self.namingTemplates['variable'].format(unit=key, feature='label')
        case_labels = ET.SubElement(mc, 'constant', attrib={'name': label_name})
        case_labels.text = value

  def _modify_inner_time_vars(self, template, case):
    """
      Modify Index var attributes of DataObjects if case._time_varname is not 'Time.'
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    # Modify dispatch groups to contain correct 'Time' and 'Year' variable.
    for group in template.find('VariableGroups').findall('Group'):
      if group.attrib['name'] in ['GRO_dispatch', 'GRO_full_dispatch_indices']:
        self._updateCommaSeperatedList(group, case.get_time_name())
        self._updateCommaSeperatedList(group, case.get_year_name())
    # Modify Data Objects to contain correct index var.
    data_objs = template.find('DataObjects')
    for index in data_objs.findall("DataSet/Index"):
      if index.get('var') == 'Time':
        index.set('var', case.get_time_name())
      if index.get('var') == 'Year':
        index.set('var', case.get_year_name())

  def _modify_inner_runinfo(self, template, case):
    """
      Defines modifications to the RunInfo of inner.xml RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
       @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    case_name = self.namingTemplates['jobname'].format(case=case.name, io='i')
    run_info = template.find('RunInfo')
    run_info.find('JobName').text = case_name
    run_info.find('WorkingDir').text = case_name
    if case.debug['enabled']:
      # need to "write full" as part of sequence, after arma sampling
      self._updateCommaSeperatedList(run_info.find('Sequence'), 'write_full', after='arma_sampling')
    # parallel
    if case.innerParallel:
      run_info.append(xmlUtils.newNode('internalParallel', text='True'))
      run_info.find('batchSize').text = f'{case.innerParallel}'

  def _modify_inner_sources(self, template, case, components, sources):
    """
      Defines modifications to the inner.xml RAVEN input file due to Sources/Placeholders.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    # for every ARMA SOURCE, we need to:
    #  - load the source to a Model
    #  - sample it N times (N = # denoise)
    #  - do signal preprocessing and dispatch
    #  - dump the results to file? --> only for viewing, debugging, so probably yes
    for source in sources:
      if source.is_type('ARMA'):
        # add a step to load the model
        self._iostep_load_rom(template, case, components, source)
        # add a step to print the rom meta
        self._iostep_rom_meta(template, source)
        # add the source to the arma-and-dispatch ensemble
        self._add_arma_to_ensemble(template, source)
        # NOTE assuming input to all ARMAs is "scaling" constant = 1.0, already in MC sampler
        if source.eval_mode == 'clustered':
          # add _ROM_Cluster to the variable group if it isn't there already
          var_group = template.find("VariableGroups/Group")
          if self.namingTemplates['cluster_index'] not in var_group.text:
            var_group.text += f",{self.namingTemplates['cluster_index']}"
          # make sure _ROM_Cluster is part of dispatch targetevaluation
          found = False
          for dataObj in template.find('DataObjects').findall('DataSet'):
            if dataObj.attrib['name'] == 'dispatch_eval':
              dispatch_eval = dataObj
              for idx in dataObj.findall('Index'):
                if idx.attrib['var'] == self.namingTemplates['cluster_index']:
                  found = True
                  break
              break
          else:
            raise RuntimeError
          if not found:
            dispatch_eval.append(xmlUtils.newNode('Index',
                                                  attrib={'var': self.namingTemplates['cluster_index']},
                                                  text='GRO_dispatch_in_Time'))

      elif source.is_type('Function'):
        # nothing to do ... ?
        pass

  def _modify_inner_components(self, template, case, components):
    """
      Defines modifications to the inner.xml RAVEN input file due to Components.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ Out, None
    """
    mc = template.find('Samplers').find('MonteCarlo')
    # find specific variable groups
    groups = {}
    var_groups = template.find('VariableGroups')
    for tag in ['capacities', 'init_disp', 'full_dispatch']:
      groups[tag] = var_groups.find(f".//Group[@name='GRO_{tag}']")

    # change inner input due to components requested
    for component in components:
      name = component.name
      # treat capacity
      ## we just need to make sure everything we need gets into the dispatch ensemble model.
      ## For each interaction of each component, that means making sure the Function, ARMA, or constant makes it.
      ## Constants from outer (namely sweep/opt capacities) are set in the MC Sampler from the outer
      ## The Dispatch needs info from the Outer to know which capacity to use, so we can't pass it from here.
      capacity = component.get_capacity(None, raw=True)
      interaction = component.get_interaction()
      parametric = capacity.is_parametric()

      if parametric:
        # this capacity is being [swept or optimized in outer] (list) or is constant (float)
        # -> so add a node, put either the const value or a dummy in place
        cap_name = self.namingTemplates['variable'].format(unit=name, feature='capacity')
        values = capacity.get_value(debug=case.debug['enabled'])
        if isinstance(values, list):
          cap_val = 42 # placeholder
        else:
          cap_val = values
        mc.append(xmlUtils.newNode('constant', attrib={'name': cap_name}, text=cap_val))
        # add component to applicable variable groups
        self._updateCommaSeperatedList(groups['capacities'], cap_name)
      elif capacity.type in ['SyntheticHistory', 'Function', 'Variable']:
        # capacity is limited by a signal, so it has to be handled in the dispatch; don't include it here.
        # OR capacity is limited by a function, and we also can't handle it here, but in the dispatch.
        pass
      else:
        raise NotImplementedError(f'Capacity from "{capacity}" not implemented yet. Component: {cap_name}')

      for tracker in component.get_tracking_vars():
        for resource in component.get_resources():
          var_name = self.namingTemplates['dispatch'].format(component=name, tracker=tracker, resource=resource)
          self._updateCommaSeperatedList(groups['init_disp'], var_name)
          self._updateCommaSeperatedList(groups['full_dispatch'], var_name)

  def _modify_inner_debug(self, template, case, components):
    """
      Modify template to work in a debug mode.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ Out, None
    """
    # RunInfo
    seq = template.find('.//RunInfo/Sequence')
    seq.text = (','.join(seq.text.split(',')[:-2])) # exclude basic stats parts
    # Steps
    for step in template.find('Steps'):
      if step.get('name') == 'arma_sampling':
        step.append(self._assemblerNode('Output', 'DataObjects', 'DataSet', 'disp_full'))
      # elif step.get('name') == 'write_summary':
      #   step.find('Output').text = 'disp_full'
    # Model
    extmod_vars = template.find('Models').find('ExternalModel').find('variables')
    self._updateCommaSeperatedList(extmod_vars, 'GRO_full_dispatch')
    self._updateCommaSeperatedList(extmod_vars, 'GRO_full_dispatch_indices')
    # DataObject
    datasets = template.find('DataObjects').findall('DataSet')
    for ds in datasets:
      if ds.attrib['name'] == 'dispatch_eval':
        break
    else:
      raise RuntimeError
    ds.append(xmlUtils.newNode('Output', text='GRO_full_dispatch'))
    for idx in ds.findall('Index'):
      self._updateCommaSeperatedList(idx, 'GRO_full_dispatch')

  def _modify_inner_optimization_settings(self, template, case):
    """
      Modifies template to include optimization settings
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    # TODO currently only modifies if optimization settings has metric and/or type, add additional settings?
    # only modify if the mode is 'opt' and <optimization_settings> has anything to modify
    if (case.get_mode() == 'opt') and (case._optimization_settings is not None):
      # optimization objective name provided (or 'missing')
      new_objective = self._build_opt_metric_out_name(case)
      # add optimization objective name to VariableGroups 'GRO_final_return' if not already there
      group = template.find('VariableGroups').find(".//Group[@name='GRO_final_return']")
      if group.text is None:
        self._updateCommaSeperatedList(group, new_objective, postion=0)
      elif (new_objective != 'missing') and (new_objective not in group.text):
        self._updateCommaSeperatedList(group, new_objective, position=0)
      # add optimization objective to PostProcessor list if not already there
      pp_node = template.find('Models').find(".//PostProcessor[@name='statistics']")
      if new_objective != 'missing':
        raven_metric_name = case._optimization_settings['metric']['name']
        prefix = case.metrics_mapping[raven_metric_name]['prefix']
        if pp_node.find(raven_metric_name) is None:
          # add subnode to PostProcessor
          if 'threshold' in case._optimization_settings['metric']:
            if raven_metric_name in ['valueAtRisk', 'expectedShortfall']:
              threshold = str(case._optimization_settings['metric']['threshold'])
            else:
              threshold = case._optimization_settings['metric']['threshold']
            new_node = xmlUtils.newNode(raven_metric_name, text=case._metric,
                                        attrib={'prefix': prefix,
                                                'threshold': threshold})
          elif 'percent' in case._optimization_settings['metric']:
            percent = str(case._optimization_settings['metric']['percent'])
            new_node = xmlUtils.newNode(raven_metric_name, text=case._metric,
                                        attrib={'prefix': prefix,
                                                'percent': percent})
          else:
            new_node = xmlUtils.newNode(raven_metric_name, text=case._metric,
                                        attrib={'prefix': prefix})
          pp_node.append(new_node)
        else:
          # check that subnode has correct values
          subnode = pp_node.find(raven_metric_name)
          # check that prefix is correct
          if prefix != subnode.attrib['prefix']:
            subnode.attrib['prefix'] = prefix
          # percentile has additional parameter to check
          if 'percent' in case._optimization_settings['metric']:
            # see if percentile already has what we need
            if str(int(case._optimization_settings['metric']['percent'])) not in subnode.attrib['percent']:
              # nope, need to add the percent to the existing attribute
              subnode.attrib['percent'] += ','+str(case._optimization_settings['metric']['percent'])
          if 'threshold' in case._optimization_settings['metric']:
            # see if the threshold is already there
            if str(case._optimization_settings['metric']['threshold']) not in subnode.attrib['threshold']:
              # nope, need to add the threshold to existing attribute
              subnode.attrib['threshold'] += ','+str(case._optimization_settings['metric']['threshold'])
      else:
        # new_objective is missing, use mean_metric
        if pp_node.find('expectedValue') is None:
          pp_node.append(xmlUtils.newNode('expectedValue', text=case._metric,
                                          attrib={'prefix': 'mean'}))
        else:
          # check that the subnode has the correct values
          subnode = pp_node.find('expectedValue')
          if 'mean' != subnode.attrib['prefix']:
            subnode.attrib['prefix'] = 'mean'
    # if no optimization settings specified, make sure mean_metric is in PostProcessor node
    elif case.get_mode() == 'opt':
      pp_node = template.find('Models').find(".//PostProcessor[@name='statistics']")
      if pp_node.find('expectedValue') is None:
        pp_node.append(xmlUtils.newNode('expectedValue', text=case._metric,
                                        attrib={'prefix': 'mean'}))
      else:
        # check that the subnode has the correct values
        subnode = pp_node.find('expectedValue')
        if 'mean' != subnode.attrib['prefix']:
          subnode.attrib['prefix'] = 'mean'

  def _modify_inner_result_statistics(self, template, case):
    """
      Modifies template to include result statistics
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    # handle VariableGroups
    var_groups = template.find('VariableGroups')
    # final return variable group (sent to outer)
    group_final_return = var_groups.find(".//Group[@name='GRO_final_return']")
    # add required defaults
    default_stats = [f'mean_{case._metric}', f'std_{case._metric}', f'med_{case._metric}']
    for stat in default_stats:
      self._updateCommaSeperatedList(group_final_return, stat)
    # make sure user provided statistics beyond defaults get there
    if any(stat not in ['expectedValue', 'sigma', 'median'] for stat in case._result_statistics):
      stats_list = self._build_result_statistic_names(case)
      for stat_name in stats_list:
        if stat_name not in default_stats:
          self._updateCommaSeperatedList(group_final_return, stat_name)
    # sweep mode has default variable names
    elif case.get_mode() == 'sweep':
      sweep_default = [f'mean_{case._metric}', f'std_{case._metric}', f'med_{case._metric}',
                       f'max_{case._metric}', f'min_{case._metric}', f'perc_5_{case._metric}',
                       f'perc_95_{case._metric}', f'samp_{case._metric}', f'var_{case._metric}']
      for sweep_name in sweep_default:
        if sweep_name not in default_stats:
          self._updateCommaSeperatedList(group_final_return, sweep_name)
    # opt mode uses optimization variable if no other stats are given, this is handled below
    if (case.get_mode == 'opt') and (case._optimization_settings is not None):
      new_metric_opt_results = self._build_opt_metric_out_name(case)
      if (new_metric_opt_results != 'missing') and (new_metric_opt_results not in group_final_return.text):
        # additional results statistics have been requested, add this metric if not already present
        self._updateCommaSeperatedList(group_final_return, new_metric_opt_results, position=0)

    # fill out PostProcessor nodes
    pp_node = template.find('Models').find(".//PostProcessor[@name='statistics']")
    # add default statistics
    stats = ['expectedValue', 'sigma', 'median']
    prefixes = ['mean', 'std', 'med']
    for stat, pref in zip(stats, prefixes):
      pp_node.append(xmlUtils.newNode(stat, text=case._metric, attrib={'prefix': pref}))
    # add any user supplied statistics beyond defaults
    if any(stat not in ['expectedValue', 'sigma', 'median'] for stat in case._result_statistics):
      for raven_metric_name in case._result_statistics:
        if raven_metric_name not in stats:
          prefix = case.metrics_mapping[raven_metric_name]['prefix']
          # add subnode to PostProcessor
          if raven_metric_name == 'percentile':
            # add percent attribute
            percent = case._result_statistics[raven_metric_name]
            if isinstance(percent, list):
              for p in percent:
                pp_node.append(xmlUtils.newNode(raven_metric_name, text=case._metric,
                                                attrib={'prefix': prefix,
                                                        'percent': p}))
            else:
              pp_node.append(xmlUtils.newNode(raven_metric_name, text=case._metric,
                                              attrib={'prefix': prefix,
                                                      'percent': percent}))
          elif raven_metric_name in ['valueAtRisk', 'expectedShortfall', 'sortinoRatio', 'gainLossRatio']:
            threshold = case._result_statistics[raven_metric_name]
            if isinstance(threshold, list):
              for t in threshold:
                pp_node.append(xmlUtils.newNode(raven_metric_name, text=case._metric,
                                                attrib={'prefix': prefix,
                                                        'threshold': t}))
            else:
              pp_node.append(xmlUtils.newNode(raven_metric_name, text=case._metric,
                                              attrib={'prefix': prefix,
                                                      'threshold': threshold}))
          else:
            pp_node.append(xmlUtils.newNode(raven_metric_name, text=case._metric,
                                            attrib={'prefix': prefix}))
    # if not specified, "sweep" mode has additional defaults
    elif case.get_mode() == 'sweep':
      stats = ['maximum', 'minimum', 'percentile', 'samples', 'variance']
      prefixes = ['max', 'min', 'perc', 'samp', 'var']
      for stat, pref in zip(stats, prefixes):
        pp_node.append(xmlUtils.newNode(stat, text=case._metric, attrib={'prefix': pref}))
    # if not specified, "opt" mode is handled in _modify_inner_optimization_settings

  def _modify_inner_data_handling(self, template, case):
    """
      Modifies template to include data handling options
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    # inner to outer
    ## default is netCDF, and is how the templates are already set up
    if case.data_handling['inner_to_outer'] == 'csv':
      # change the output IOStep outstream to do CSV instead of database
      print(
      template.find('Steps').find(".//IOStep[@name='database']")
      )
      db = template.find('Steps').find('.//IOStep[@name="database"]').find('.//Output[@class="Databases"]')
      db.attrib.update({'class': 'OutStreams', 'type': 'Print'})
      # the database and outstream print have the same name, so don't need to change text of node


  ##### CASHFLOW #####
  def _modify_cash(self, template, case, components, sources):
    """
      Defines modifications to the cash.xml extension to RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, template, xml.etree.ElementTree.Element, modified template
    """
    self._modify_cash_Global(template, case)
    self._modify_cash_components(template, case, components)
    return template

  def _modify_cash_Global(self, template, case):
    """
      Defines modifications to Global of cash.xml extension to RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ Out, None
    """
    # load variables
    tax = case._global_econ['tax']
    verbosity = case._global_econ['verbosity']
    inflation = case._global_econ['inflation']
    indicator = case._global_econ['Indicator']
    discountRate = case._global_econ['DiscountRate']
    projectTime = case._global_econ.get('ProjectTime', None)
    # set variables
    template.attrib['verbosity'] = str(verbosity)
    cash_global = template.find('Global')
    cash_global.find('DiscountRate').text = str(discountRate)
    cash_global.find('tax').text = str(tax)
    cash_global.find('inflation').text = str(inflation)
    cash_global.find('Indicator').attrib['name'] = indicator['name'][0]
    cash_global.find('Indicator').text = '\n      '.join(indicator['active'][:])
    if projectTime is not None:
      cash_global.append(xmlUtils.newNode('ProjectTime', text=str(projectTime)))

  def _modify_cash_components(self, template, case, components):
    """
      Defines modifications to Components of cash.xml extension to RAVEN input file.
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ Out, template, xml.etree.ElementTree.Element, modified template
    """
    for component in components:
      subComp = xmlUtils.newNode('Component', attrib={'name': component.name}, text='')
      subEconomics = component.get_economics()
      Life_time = subEconomics._lifetime
      subComp.append(xmlUtils.newNode('Life_time', text=str(Life_time)))
      cfs=xmlUtils.newNode('CashFlows')
      for subCash in subEconomics._cash_flows:
        driverName  = self.namingTemplates['cashfname'].format(component=subCash._component.name, cashname=subCash._driver.name)
        driverType   = subCash._driver.type

        inflation    = subCash._inflation
        mult_target  = subCash._mult_target
        name         = subCash.name
        tax          = subCash._taxable
        depreciation = subCash._depreciate
        if subCash._type == 'one-time':
          cfNode =  xmlUtils.newNode('Capex', text='', attrib={'name': f'{name}',
                                                                'tax':tax,
                                                                'inflation': inflation,
                                                                'mult_target': mult_target
                                                                })
          cfNode.append(xmlUtils.newNode('driver',text = driverName))
          cfNode.append(xmlUtils.newNode('alpha',text = subCash._alpha.get_value()))
          cfNode.append(xmlUtils.newNode('reference',text = subCash._reference.get_value()))
          cfNode.append(xmlUtils.newNode('X',text = subCash._scale.get_value()))
          if depreciation:
            cfNode.append(xmlUtils.newNode('depreciation',attrib={'scheme':'MACRS'}, text = depreciation))
          cfs.append(cfNode)
        else:
          cfNode =  xmlUtils.newNode('Recurring', text='', attrib={'name': f'{name}',
                                                                   'tax':tax,
                                                                   'inflation': inflation,
                                                                   'mult_target': mult_target
                                                                    })
          cfNode.append(xmlUtils.newNode('driver',
          text = self.namingTemplates['re_cash'].format(period=subCash._period,
                                                        driverType = driverType,
                                                        driverName = f'_{component.name}_{name}')))

          cfNode.append(xmlUtils.newNode('alpha',text = '-1.0'))
          cfs.append(cfNode)
      subComp.append(cfs)
      template.append(subComp)



  ##### OTHER UTILS #####
  def _add_arma_to_ensemble(self, template, source):
    """
      Adds an ARMA to EnsembleModel evaluation
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, source, HERON Placeholder, information about ARMA to be used
      @ Out, None
    """
    # pre-locate some useful nodes
    ens = template.find('Models').findall('EnsembleModel')[0] # NOTE relies on position
    data_objs = template.find('DataObjects')
    # pre-write some useful names
    inp_name = self.namingTemplates['data object'].format(source=source.name, contents='placeholder')
    eval_name = self.namingTemplates['data object'].format(source=source.name, contents='samples')
    # variables that are outputs of the rom
    out_vars = source.get_variable()

    # model was added during adding the loading step, so no need to add it to Models block
    # we do need to add it to the ensemble model though
    new_model = self._assemblerNode('Model', 'Models', 'ROM', source.name)
    new_model.append(self._assemblerNode('Input', 'DataObjects', 'PointSet', inp_name))
    new_model.append(self._assemblerNode('TargetEvaluation', 'DataObjects', 'DataSet', eval_name))
    ens.append(new_model)

    # create the data objects
    deps = {self.__case.get_time_name(): out_vars,
            self.__case.get_year_name(): out_vars}
    if source.eval_mode == 'clustered':
      deps[self.namingTemplates['cluster_index']] = out_vars

    self._create_dataobject(data_objs, 'PointSet', inp_name, inputs=['scaling'])
    self._create_dataobject(data_objs, 'DataSet', eval_name,
                            inputs=['scaling'],
                            outputs=out_vars,
                            depends=deps)

    # add variables to dispatch input requirements
    ## before all else fails, use variable groups
    # find dispatch_in_time group
    for group in (g for g in template.find('VariableGroups') if (g.tag == 'Group')):
      if group.attrib['name'] == 'GRO_dispatch_in_Time':
        break
    else:
      raise RuntimeError
    for var in out_vars:
      self._updateCommaSeperatedList(group, var)

  def _create_dataobject(self, dataobjects, typ, name, inputs=None, outputs=None, depends=None):
    """
      Creates a data object candidate to go to base class
      @ In, dataobjects, xml.etree.ElementTreeElement, DataObjects node
      @ In, typ, str, type of data object
      @ In, name, str, name of data object
      @ In, inputs, list(str), optional, input variable names
      @ In, outputs, list(str), optional, output variable names
      @ In, depends, dict, optional, time-dependency as {index: [stuff that depends on index]}
      @ Out, None
    """
    assert typ in ['PointSet', 'HistorySet', 'DataSet']
    new = xmlUtils.newNode(typ, attrib={'name':name})
    if inputs is not None:
      new.append(xmlUtils.newNode('Input', text=','.join(inputs)))
    if outputs is not None:
      new.append(xmlUtils.newNode('Output', text=','.join(outputs)))
    # index dependence
    ## if pointset, you went wrong somewhere if you gave dependencies
    if typ == 'PointSet':
      assert depends is None
    ## if a history set, there better only be one
    elif typ == 'HistorySet':
      assert depends is not None
      assert len(depends) == 1, f'Depends is: {depends}'
      opt = xmlUtils.newNode('options')
      opt.append(xmlUtils.newNode('pivotParameter', text=list(depends.keys())[0]))
      new.append(opt)
    ## otherwise, if dataset, follow the dependencies
    elif typ == 'DataSet':
      if depends is not None:
        for index, dep in depends.items():
          assert isinstance(dep, list)
          new.append(xmlUtils.newNode('Index', attrib={'var':index}, text=', '.join(dep)))
    dataobjects.append(new)

  def _iostep_load_rom(self, template, case, components, source):
    """
      for INNER, creates new IOStep for loading a ROM
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, case, HERON Case, defining Case instance
      @ In, components, list, list of HERON Component instances for this run
      @ In, sources, list, list of HERON Placeholder instances for this run
      @ Out, None
    """
    rom_name = source.name
    rom_source = source._target_file
    # add the step itself
    step_name = self.namingTemplates['stepname'].format(action='read', subject=rom_name)
    new_step = xmlUtils.newNode('IOStep', attrib={'name': step_name})
    new_step.append(self._assemblerNode('Input', 'Files', '', rom_source))
    new_step.append(self._assemblerNode('Output', 'Models', 'ROM', rom_name))
    template.find('Steps').append(new_step)
    # update the sequence
    self._updateCommaSeperatedList(template.find('RunInfo').find('Sequence'), new_step.attrib['name'], position=0)
    # add the model
    model = xmlUtils.newNode('ROM', attrib={'name':rom_name, 'subType':'pickledROM'})
    econ_comps = list(comp.get_economics() for comp in components)
    econ_global_params = case.get_econ(econ_comps)
    econ_global_settings = CashFlows.GlobalSettings()
    econ_global_settings.setParams(econ_global_params)
    ## update the ARMA model to sample a number of years equal to the ProjectLife from CashFlow
    if source.needs_multiyear is not None:
      multiyear = xmlUtils.newNode('Multicycle')
      multiyear.append(xmlUtils.newNode('cycles', text=source.needs_multiyear))
      model.append(multiyear)
    if source.limit_interp is not None:
      maxCycles = model.find('maxCycles')
      if maxCycles is not None:
        maxCycles.text = source.limit_interp
      else:
        model.append(xmlUtils.newNode('maxCycles', text=source.limit_interp))
    # change eval mode?
    if source.eval_mode == 'clustered':
      model.append(xmlUtils.newNode('clusterEvalMode', text='clustered'))
    template.find('Models').append(model)
    # add a file
    ## NOTE: the '..' assumes there is a working dir that is not ".", which should always be true.
    ## ALSO NOTE: the path to the ARMA file should now be absolute, so no directory fiddling necessary?
    template.find('Files').append(xmlUtils.newNode('Input', attrib={'name':rom_source}, text=rom_source)) # '../../'+rom_source
    # done

  def _iostep_rom_meta(self, template, source):
    """
      for INNER, create an IOStep for printing the ROM meta
      @ In, template, xml.etree.ElementTree.Element, root of XML to modify
      @ In, source, HERON Placeholder, instance to add rom meta use for
      @ Out, None
    """
    rom_name = source.name
    # create the output data object
    objs = template.find('DataObjects')
    obj_name = f'{rom_name}_meta'
    self._create_dataobject(objs, 'DataSet', obj_name)
    # create the output outstream
    os_name = obj_name
    streams = template.find('OutStreams')
    if streams is None:
      streams = xmlUtils.newNode('OutStreams')
      template.append(streams)
    new = xmlUtils.newNode('Print', attrib={'name': os_name})
    new.append(xmlUtils.newNode('type', text='csv'))
    new.append(xmlUtils.newNode('source', text=obj_name))
    streams.append(new)
    # create the step
    step_name = self.namingTemplates['stepname'].format(action='print_meta', subject=rom_name)
    new_step = xmlUtils.newNode('IOStep', attrib={'name': step_name})
    new_step.append(self._assemblerNode('Input', 'Models', 'ROM', rom_name))
    new_step.append(self._assemblerNode('Output', 'DataObjects', 'PointSet', obj_name))
    new_step.append(self._assemblerNode('Output', 'OutStreams', 'Print', os_name))
    template.find('Steps').append(new_step)
    self._updateCommaSeperatedList(template.find('RunInfo').find('Sequence'), step_name, position=1)

  def _remove_by_name(self, root, removable):
    """
      Removes subs of "root" whose "name" attribute is in "removable"
      @ In, root. ET.Element, node whose subs should be searched through
      @ In, removable, list(str), names to remove
      @ Out, None
    """
    to_remove = []
    for node in root:
      if node.get('name', None) in removable:
        to_remove.append(node)
    for node in to_remove:
      root.remove(node)

  def _build_opt_metric_out_name(self, case):
    """
      Constructs the output name of the metric specified as the optimization objective
      @ In, case, HERON Case, defining Case instance
      @ Out, opt_out_metric_name, str, output metric name for use in inner/outer files
    """
    try:
      # metric name in RAVEN
      metric_raven_name = case._optimization_settings['metric']['name']
      # potential metric name to add
      opt_out_metric_name = case.metrics_mapping[metric_raven_name]['prefix']
      # do I need to add a percent or threshold to this name?
      if metric_raven_name == 'percentile':
        opt_out_metric_name += '_' + str(case._optimization_settings['metric']['percent'])
      elif metric_raven_name in ['valueAtRisk', 'expectedShortfall', 'sortinoRatio', 'gainLossRatio']:
        opt_out_metric_name += '_' + str(case._optimization_settings['metric']['threshold'])
      opt_out_metric_name += '_'+case._metric
    except (TypeError, KeyError):
      # <optimization_settings> node not in input file OR
      # 'metric' is missing from _optimization_settings
      opt_out_metric_name = 'missing'

    return opt_out_metric_name

  def _build_result_statistic_names(self, case):
    """
      Constructs the names of the statistics requested for output
      @ In, case, HERON Case, defining Case instance
      @ Out, names, list, list of names of statistics requested for output
    """
    names = []
    for name in case._result_statistics:
      out_name = case.metrics_mapping[name]['prefix']
      # do I need to add percent or threshold?
      if name in ['percentile', 'valueAtRisk', 'expectedShortfall', 'sortinoRatio', 'gainLossRatio']:
        # multiple percents or thresholds may be specified
        if isinstance(case._result_statistics[name], list):
          for attrib in case._result_statistics[name]:
            names.append(out_name+'_'+attrib+'_'+case._metric)
        else:
          names.append(out_name+'_'+case._result_statistics[name]+'_'+case._metric)
      else:
        out_name += '_'+case._metric
        names.append(out_name)

    return names




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Class for managing interactions with the Dispatchers.
"""


import os
import sys
import pickle as pk
from time import time as run_clock

import numpy as np
from typing_extensions import final

from . import _utils as hutils
from . import SerializationManager

try:
  from ravenframework.PluginBaseClasses.ExternalModelPluginBase import ExternalModelPluginBase
  import TEAL
except ModuleNotFoundError:
  raven_path = hutils.get_raven_loc()
  sys.path.append(raven_path)
  from ravenframework.PluginBaseClasses.ExternalModelPluginBase import ExternalModelPluginBase
  sys.path.pop()

  cashflow_path = os.path.abspath(os.path.join(hutils.get_cashflow_loc(raven_path=raven_path), '..'))
  sys.path.append(cashflow_path)
  import TEAL
from TEAL.src import CashFlows
from TEAL.src.main import run as CashFlow_run

# make functions findable
sys.path.append(os.getcwd())

class DispatchRunner:
  """
    Manages the interface between RAVEN and running the dispatch
  """
  # TODO move naming templates to a common place for consistency!
  naming_template = {
    'comp capacity': '{comp}_capacity',
    'dispatch var': 'Dispatch__{comp}__{tracker}__{res}',
  }

  def __init__(self):
    """
      Constructor. Note instances of this are tied to single ExternalModel runs.
      @ In, None
      @ Out, None
    """
    self._case = None              # HERON case
    self._components = None        # HERON components list
    self._sources = None           # HERON sources (placeholders) list
    self._override_time = None     # override for micro parameter
    self._save_dispatch = False    # if True then maintain and return full dispatch record

  #####################
  # API
  def override_time(self, new):
    """
      Sets the micro time-like parameter for dispatch optimization.
      @ In, new, list(int), arguments as for numpy linpace.
      @ Out, None
    """
    self._override_time = new

  def load_heron_lib(self, path):
    """
      Loads HERON objects from library file.
      @ In, path, str, path (including filename) to HERON library
      @ Out, None
    """
    case, components, sources = SerializationManager.load_heron_lib(path, retry=6)
    # arguments
    self._case = case              # HERON case
    self._components = components  # HERON components list
    self._sources = sources        # HERON sources (placeholders) list
    # derivative
    self._dispatcher = self._case.dispatcher
    if self._case.debug['enabled']:
      self._save_dispatch = True

  def extract_variables(self, raven, raven_dict):
    """
      Extract variables from RAVEN and apply them to HERON objects
      @ In, raven, object, RAVEN external model object
      @ In, raven_dict, dict, RAVEN input dictionary
      @ Out, pass_vars, dict, variables to pass to dispatcher
    """
    pass_vars = {}
    history_structure = {}
    # investigate sources for required ARMA/CSV information
    for source in self._sources:
      if source.is_type('ARMA') or source.is_type("CSV"):
        # get structure of ARMA/CSV
        vars_needed = source.get_variable()
        for v in vars_needed:
          pass_vars[v] = getattr(raven, v)

    # get the key to mapping RAVEN multidimensional variables
    if hasattr(raven, '_indexMap'):
      pass_vars['_indexMap'] = raven._indexMap[0] # 0 is only because of how RAVEN EnsembleModel handles variables
      # collect all indices # TODO limit to those needed by sources?
      for target, required_indices in pass_vars['_indexMap'].items():
        for index in filter(lambda idx: idx not in pass_vars, required_indices):
          pass_vars[index] = getattr(raven, index)
    else:
      # NOTE this should ONLY BE POSSIBLE if no ARMAs or CSVs are in use!
      pass

    # variable for "time" discretization, if present
    year_var = self._case.get_year_name()
    time_var = self._case.get_time_name()
    time_vals = getattr(raven, time_var, None)
    if time_vals is not None:
      pass_vars[time_var] = time_vals

    # TODO references to all ValuedParams should probably be registered somewhere
    # like maybe in the VPFactory, then we can loop through and look for info
    # that we know from Outer and fill in the blanks? Maybe?
    for magic in self._case.dispatch_vars.keys():
      val = getattr(raven, f'{magic}_dispatch', None)
      if val is not None:
        pass_vars[magic] = float(val)

    # component capacities
    for comp in self._components:
      name = self.naming_template['comp capacity'].format(comp=comp.name)
      update_capacity = raven_dict.get(name) # TODO is this ever not provided?
      if update_capacity is not None:
        comp.set_capacity(update_capacity)
        pass_vars[f'{comp.name}_capacity'] = update_capacity
    # TODO other case, component properties

    # check macro parameter
    if year_var in dir(raven):
      year_vals = getattr(raven, year_var)
      year_size = year_vals.size
      project_life = hutils.get_project_lifetime(self._case, self._components) - 1 # 1 for construction year
      if year_size != project_life:
        raise RuntimeError(f'Provided macro variable "{year_var}" is length {year_size}, ' +
                           f'but expected project life is {project_life}! ' +
                           f'"{year_var}" values: {year_vals}')

    # load ARMA signals
    for source in self._sources:
      if source.is_type('ARMA'):
        vars_needed = source.get_variable()
        for v in vars_needed:
          vals = getattr(raven, v, None)
          # checks
          if vals is None:
            raise RuntimeError(f'HERON: Expected ARMA variable "{v}" was not passed to DispatchManager!')
          pass_vars[v] = vals
    return pass_vars

  def run(self, raven_vars):
    """
      Runs dispatcher.
      @ In, raven_vars, dict, dictionary of variables from raven to pass through
      @ Out, all_dispatch, DispatchState, results of dispatching
      @ Out, metrics, dict, economic metric results
    """
    ## TODO move as much of this as possible to "intialize" instead of "run"!
    # build meta variable
    ## this will be passed to external functions
    print("RUNNING HERON DISPATCH MANAGER")
    heron_meta = {}
    heron_meta['Case'] = self._case
    heron_meta['Components'] = self._components
    heron_meta['Sources'] = self._sources
    heron_meta['RAVEN_vars_full'] = raven_vars
    # build indexer for components
    ## indexer is as {component: {res: index}} where index is a standardized index for tracking activity
    heron_meta['resource_indexer'] = dict((comp, dict((res, r) for r, res in enumerate(comp.get_resources())))
                                          for comp in self._components)
    # store meta
    meta = {'HERON': heron_meta}
    # do some checking a priori
    ## TODO can we do any of this at compile time instead of run time?
    ## -> probably, by investigating the ARMA file to be used
    self._check_time(raven_vars)
    self._check_signals(raven_vars)
    # determine analysis structure
    all_structure = self._get_structure(raven_vars)
    # just need the summary info for now
    structure = all_structure['summary']
    # set up evaluation loops
    if structure['clusters']:
      seg_type = 'Cluster'
      segs = structure['clusters']
    elif structure['segments']:
      seg_type = 'Segment'
      segs = structure['Segment']
    else:
      seg_type = 'All'
      segs = range(1)
    interp_years = range(*structure['interpolated'])
    project_life = hutils.get_project_lifetime(self._case, self._components) - 1 # 1 for construction year
    # if the ARMA is a single year, no problem, we replay it for each year
    # if the ARMA is the same or greater number of years than the project_life, we can use the ARMA still
    # otherwise, there's a problem
    if 1 < len(interp_years) < project_life:
      raise IOError(f'An interpolated ARMA ROM was used, but there are less interpolated years ' +
                    f'({list(range(*structure["interpolated"]))}) ' +
                    f'than requested project years ({project_life})!')
    # do the dispatching
    all_dispatch, metrics = self._do_dispatch(meta, all_structure, project_life, interp_years, segs, seg_type)
    return all_dispatch, metrics

  def save_variables(self, raven, all_dispatch, metrics):
    """
      generates RAVEN-acceptable variables
      Saves variables on "raven" object for returning
      @ In, raven, object, RAVEN object for setting values
      @ In, all_dispatch, dict, dispatch values
      @ In, metrics, dict, economic metrics
    """
    template = self.naming_template['dispatch var']
    for y, (year, year_data) in enumerate(all_dispatch.items()):
      for c, (cluster, dispatch) in enumerate(year_data.items()):
        dispatches = dispatch.create_raven_vars(template)
        # set up index map, first time only
        if y == c == 0:
          # string names
          year_name = self._case.get_year_name()
          clst_name = '_ROM_Cluster'
          time_name = self._case.get_time_name()
          # number of entries for each dim
          n_year = len(all_dispatch)
          n_clst = len(year_data)
          n_time = len(dispatch._times) # NOTE assuming same across clusters!
          # set indices on raven
          setattr(raven, time_name, np.asarray(dispatch._times))
          setattr(raven, year_name, np.asarray(list(all_dispatch.keys())))
          setattr(raven, clst_name, np.arange(n_clst))
          if not getattr(raven, '_indexMap', None):
            raven._indexMap = np.atleast_1d({})
        for var_name, data in dispatches.items():
          # if first time, initialize data structure
          if y == c == 0:
            shape = (n_year, n_clst, n_time)
            setattr(raven, var_name, np.empty(shape)) # NOTE could use np.zeros, but slower?
          getattr(raven, var_name)[y, c] = data
          getattr(raven, '_indexMap')[0][var_name] = [year_name, clst_name, time_name]
    for metric, value in metrics.items():
      setattr(raven, metric, np.atleast_1d(value))
    # if component capacities weren't given by Outer, save them as part of Inner
    for comp in self._components:
      cap_name = self.naming_template['comp capacity'].format(comp=comp.name)
      if cap_name not in dir(raven):
        # TODO what value should actually be used?
        setattr(raven, cap_name, -42)

  #####################
  # UTILITIES
  def _do_dispatch(self, meta, all_structure, project_life, interp_years, segs, seg_type):
    """
      perform dispatching
      @ In, meta, dict, dictionary of passthrough variables
      @ In, project_life, int, total analysis years (e.g. 30)
      @ In, interp_years, list, actual analysis tagged years (e.g. range(2015, 2045))
      @ In, segs, list(int), segments/clusters/divisions
      @ In, seg_type, str, "segment" or "cluster" if segmented or clustered
      @ Out, dispatch_results, list(DispatchState), results of dispatch for each segment/cluster and year
    """
    structure = all_structure['summary']
    ## FINAL settings/components/cashflows use the multiplicity of divisions for aggregated evaluation
    final_settings, final_components = self._build_econ_objects(self._case, self._components, project_life)
    active_index = {}
    dispatch_results = {}
    yearly_cluster_data = next(iter(all_structure['details'].values()))['clusters']
    for year in range(project_life):
      interp_year = interp_years[year] if len(interp_years) > 1 else (interp_years[0] + year)
      if self._save_dispatch:
        dispatch_results[interp_year] = {}
      # If the ARMA is interpolated, we need to track which year we're in.
      # Otherwise, use just the nominal first year.
      active_index['year'] = year if len(range(*structure['interpolated'])) > 1 else 0 # FIXME MacroID not year
      for s, seg in enumerate(segs):
        multiplicity = self._update_meta_for_segment(meta, seg, interp_year, yearly_cluster_data,
                                                     interp_years, active_index, all_structure)
        # perform dispatch
        dispatch = self._dispatcher.dispatch(self._case, self._components, self._sources, meta)
        if self._save_dispatch:
          dispatch_results[interp_year][seg] = dispatch
        # build evaluation cash flows
        self._segment_cashflow(meta, s, seg, year, dispatch, multiplicity,
                               project_life, interp_years, all_structure, final_components)
    # TEAL, take it away.
    cf_metrics = self._final_cashflow(meta, final_components, final_settings)
    return dispatch_results, cf_metrics

  def _build_econ_objects(self, heron_case, heron_components, project_life):
    """
      Generates CashFlow.CashFlow instances from HERON CashFlow instances
      Note the only reason there's a difference is because HERON needs to retain some level of
      flexibility in the parameter values until this method is called, whereas CashFlow expects
      them to be evaluated.
      @ In, heron_case, HERON Case instance, global HERON settings for this analysis
      @ In, heron_components, list, HERON component instances
      @ In, project_life, int, number of years to evaluate project
      @ Out, global_settings, CashFlow.GlobalSettings instance, settings for CashFlow analysis
      @ Out, cf_components, dict, CashFlow component instances
    """
    heron_econs = list(comp.get_economics() for comp in heron_components)
    # build global econ settings for CashFlow
    global_params = heron_case.get_econ(heron_econs)
    global_settings = CashFlows.GlobalSettings()
    global_settings.setParams(global_params)
    global_settings._verbosity = 0 # FIXME direct access, also make user option?
    # build CashFlow component instances
    cf_components = {}
    for c, cfg in enumerate(heron_econs):
      # cfg is the cashflowgroup connected to the heron component
      # get the associated heron component
      heron_comp = heron_components[c]
      comp_name = heron_comp.name
      # build CashFlow equivalent component
      cf_comp = CashFlows.Component()
      cf_comp_params = {'name': comp_name,
                        'Life_time': cfg.get_lifetime(),
                        # TODO StartTime, Repetitions, tax, inflation
                       }
      cf_comp.setParams(cf_comp_params)
      cf_components[comp_name] = cf_comp
      # create all the CashFlow.CashFlows (cf_cf) for the CashFlow.Component
      cf_cfs = []
      for heron_cf in cfg.get_cashflows():
        cf_name = heron_cf.name
        # the way to build it slightly changes depending on the CashFlow type
        if heron_cf._type == 'repeating': # FIXME protected access
          cf_cf = CashFlows.Recurring()
          cf_cf_params = {'name': cf_name,
                          'X': 1.0,
                          'mult_target': heron_cf._mult_target, # FIXME protected access
                          }
          cf_cf.setParams(cf_cf_params)
          cf_cf.initParams(project_life)
        elif heron_cf._type == 'one-time': # FIXME protected access
          cf_cf = CashFlows.Capex()
          cf_cf.name = cf_name
          cf_cf.initParams(cf_comp.getLifetime())
          # alpha, driver aren't known yet, so set those later
        else:
          raise NotImplementedError(f'Unknown HERON CashFlow Type: {heron_cf._type}')
        # store new object
        cf_cfs.append(cf_cf)
      cf_comp.addCashflows(cf_cfs)
    return global_settings, cf_components

  def _update_meta_for_segment(self, meta, seg, interp_year, yearly_cluster_data,
                               interp_years, active_index, all_structure) -> int:
    """
      Updates the "meta" auxiliary information variable to use info specific to the segment
      @ In, meta, dict, auxiliary information
      @ In, seg, int, id for current segment (or cluster)
      @ In, interp_year, int, identifier for active year
      @ In, yearly_cluster_data, TODO
      @ In, interp_years, TODO
      @ In, active_index, dict, active indices (including year, segment)
      @ In, all_structure, dict, structure of ARMA sample/realization
      @ Out, multiplicity, int, number of segments represented by this segment within the year
    """
    # get cluster info from the first source -> assumes all clustering is aligned!
    # -> find the info for this cluster -> FIXME this should be restructured so searching isn't necessary!
    if interp_year in yearly_cluster_data:
      clusters_info = yearly_cluster_data[interp_year]
    else:
      clusters_info = yearly_cluster_data[interp_years[0]]
    for cl_info in clusters_info:
      if cl_info['id'] == seg:
        break
    else:
      raise RuntimeError
    # how many clusters does this one represent?
    multiplicity = len(cl_info['represents'])

    # update meta for the current segment
    active_index['division'] = seg
    meta['HERON']['active_index'] = active_index
    # truncate signals to appropriate Year, Cluster
    ## -> chop up raven_vars for sources to corresponding segment/cluster, year, and time
    meta['HERON']['RAVEN_vars'] = self._slice_signals(all_structure, meta['HERON'])
    return multiplicity

  def _segment_cashflow(self, meta, s, seg, year, dispatch, multiplicity,
                        project_life, interp_years, all_structure, final_components) -> None:
    """
      Update TEAL CashFlow objects with new dispatch information for a segment
      @ In, TODO
      @ Out, None
    """
    # LOCAL component cashflows are SPECIFIC TO A DIVISION
    _, local_comps = self._build_econ_objects(self._case, self._components, project_life)
    meta['HERON']['active_index'] = {'year': year if len(interp_years) > 1 else 0, 'division': seg,}
    meta['HERON']['RAVEN_vars'] = self._slice_signals(all_structure, meta['HERON'])
    pivot_var = meta['HERON']['Case'].get_time_name()
    times = meta['HERON']['RAVEN_vars'][pivot_var]
    specific_meta = dict(meta) # TODO more deepcopy needed?
    resource_indexer = meta['HERON']['resource_indexer']
    for comp in self._components:
      # get corresponding current and final CashFlow.Component
      cf_comp = local_comps[comp.name]
      final_comp = final_components[comp.name]
      # sanity check
      if comp.name != cf_comp.name: raise RuntimeError
      specific_meta['HERON']['component'] = comp
      specific_meta['HERON']['all_activity'] = dispatch
      specific_activity = {}
      final_cashflows = final_comp.getCashflows()
      for f, heron_cf in enumerate(comp.get_cashflows()):
        # get the corresponding CashFlow.CashFlow
        cf_cf = cf_comp.getCashflows()[f]
        final_cf = final_cashflows[f]
        # sanity continued
        if not (cf_cf.name == final_cf.name == heron_cf.name):
            raise RuntimeError

        ## FIXME time then cashflow, or cashflow then time?
        ## "activity" is the same for every cashflow at a point in
        ## time, but many cashflows only need to be evaluated once
        ## and we can vectorize ...

        ## FIXME maybe "if activity is None" approach, so it gets
        ## filled on the first cashflow when looping through time.

        ## TODO we assume Capex and Recurring Year do not depend
        ## on the Activity

        if cf_cf.type == 'Capex':
          # Capex cfs should only be constructed in the first  of the project life
          # FIXME is this doing capex once per segment, or once per life?
          if year == 0 and s == 0:
            params = heron_cf.calculate_params(specific_meta) # a, D, Dp, x, cost
            cf_params = {'name': cf_cf.name,
                          'mult_target': heron_cf._mult_target,
                          'depreciate': heron_cf._depreciate,
                          'alpha': params['alpha'],
                          'driver': params['driver'],
                          'reference': params['ref_driver'],
                          'X': params['scaling'],}
            cf_cf.setParams(cf_params)

            ## Because alpha, driver, etc are only set once for capex cash flows,
            ## we can just hot swap this cashflow into the final_comp, I think ...

            ## I believe we can do this because Capex are division-independent?
            ## Can we just do it once instead of once per division?

            final_comp._cashFlows[f] = cf_cf
            # depreciators
            # FIXME do we need to know alpha, drivers first??
            if heron_cf._depreciate and cf_cf.getAmortization() is None:
              cf_cf.setAmortization('MACRS', heron_cf._depreciate)
              deprs = cf_comp._createDepreciation(cf_cf)
              final_comp._cashFlows.extend(deprs)
        elif cf_cf.type == 'Recurring':
          # yearly recurring only need setting up once per year
          if heron_cf.get_period() == 'year':
            if s == 0:
              params = heron_cf.calculate_params(specific_meta) # a, D, Dp, x, cost
              contrib = params['cost']
              final_cf._yearlyCashflow[year + 1] += contrib
          # hourly recurring need iteration over time
          elif heron_cf.get_period() == 'hour':
            for t, time in enumerate(times):
              # fill in the specific activity for this time stamp
              for track_var in comp.get_tracking_vars():
                specific_activity[track_var] = {}
                for resource, r in resource_indexer[comp].items():
                  specific_activity[track_var][resource] = dispatch.get_activity(comp, track_var, resource, time)
              specific_meta['HERON']['time_index'] = t
              specific_meta['HERON']['time_value'] = time
              # TODO does the rest need to be available?
              specific_meta['HERON']['activity'] = specific_activity
              # contribute to cashflow (using sum as discrete integral)
              # NOTE that intrayear depreciation is NOT being considered here
              params = heron_cf.calculate_params(specific_meta) # a, D, Dp, x, cost
              contrib = params['cost'] * multiplicity
              final_cf._yearlyCashflow[year+1] += contrib
          else:
            raise NotImplementedError(
                f'Unrecognized Recurring period for "{comp.name}" cashflow "{heron_cf.name}": {heron_cf.get_period()}'
            )
        else:
            raise NotImplementedError(
                f'Unrecognized CashFlow type for "{comp.name}" cashflow "{heron_cf.name}": {cf_cf.type}'
            )
        # end CashFlow type if
      # end CashFlow per Component loop
    # end Component loop

  def _final_cashflow(self, meta, final_components, final_settings) -> dict:
    """
      Perform final cashflow calculations using TEAL.
      @ In, meta, dict, auxiliary information
      @ In, final_components, list, completed TEAL component objects
      @ In, final_settings, TEAL.Settings, completed TEAL settings object
      @ Out, cf_metrics, dict, values for calculated metrics
    """
    print('****************************************')
    print('* Starting final cashflow calculations *')
    print('****************************************')
    raven_vars = meta['HERON']['RAVEN_vars_full']
    # DEBUGG
    print('DEBUGG CASHFLOWS')
    for comp_name, comp in final_components.items():
      print(f' ... comp {comp_name} ...')
      for cf in comp.getCashflows():
        print(f' ... ... cf {cf.name} ...')
        print(f' ... ... ... D', cf._driver)
        print(f' ... ... ... a', cf._alpha)
        print(f' ... ... ... Dp', cf._reference)
        print(f' ... ... ... x', cf._scale)
        if hasattr(cf, '_yearlyCashflow'):
          print(f' ... ... ... hourly', cf._yearlyCashflow)
    # END DEBUGG
    cf_metrics = CashFlow_run(final_settings, list(final_components.values()), raven_vars)
    # DEBUGG
    print('****************************************')
    print('DEBUGG final cashflow metrics:')
    for k, v in cf_metrics.items():
      print('  ', k, v)
    print('****************************************')
    # END DEBUGG
    return cf_metrics

  def _get_structure(self, raven_vars):
    """
      interpret the clustering information from the ROM
      @ In, raven_vars, dict, variables coming from RAVEN
      @ Out, all_structure, dict, structure (multiyear, cluster/segments, etc) specifications
    """
    all_structure = {'details': {}, 'summary': {}}
    found = False
    assert self._sources is not None
    for source in self._sources:
      if source.is_type("ARMA"):
        structure = hutils.get_synthhist_structure(source._target_file)
        all_structure["details"][source] = structure
        found = True
        break

    if not found:
      for source in self._sources:
        if source.is_type("CSV"):
          structure = hutils.get_csv_structure(
              source._target_file,
              self._case.get_year_name(),
              self._case.get_time_name()
          )
          all_structure['details'][source] = structure
          found = True
          break

    # It's important to note here. We do not anticipate users mixing
    # ARMA & CSV sources, we also don't account for discrepancies in
    # time-steps between CSV and ARMA. Eventually we may need to modify
    # the code to allow for mixed use and determine compatibility of
    # time-steps.
    if not found:
      raise RuntimeError('No ARMA or CSV found in sources! Temporal mapping is missing.')

    # TODO check consistency between ROMs?
    # for now, just summarize what we found -> take it from the first source
    summary_info = next(iter(all_structure['details'].values()))
    interpolated = (summary_info['macro']['first'], summary_info['macro']['last'] + 1) if 'macro' in summary_info else (0, 1)
    # further, also take cluster structure from the first year only
    first_year_clusters = next(iter(summary_info['clusters'].values())) if 'clusters' in summary_info else {}
    clusters = list(cl['id'] for cl in first_year_clusters)
    all_structure['summary'] = {'interpolated': interpolated,
                                'clusters': clusters,
                                'segments': 0, # FIXME XXX
                                'macro_info': summary_info['macro'] if 'macro' in summary_info else {},
                                'cluster_info': first_year_clusters,
                                } # TODO need to add index/representivity references!
    return all_structure

  def _check_signals(self, raven_vars):
    """
      Checks the length of histories to assure they are consistent.
      TODO may not be necessary if we can interpolate histories!
      @ In, raven_vars, dict, dictionary of variables from RAVEN
      @ Out, None
    """
    # check the "signal" history length
    # TODO this isn't strictly required; we can interpolate sample ARMAs, right?
    signal_shapes = {}
    for source in (item for item in self._sources if item.is_type('ARMA')):
      if not source.is_type('ARMA'):
        continue
      var_names = source.get_variable()
      if var_names is None:
        var_names = []
      for name in var_names:
        if name not in signal_shapes:
          s = raven_vars[name].shape
          signal_shapes[name] = s
    if not all((shape == s) for shape in signal_shapes.values()):
      print('History shapes:', signal_shapes)
      print('Index Map:', raven_vars.get('_indexMap', None))
      raise IOError('Synthetic histories are not of consistent shape! See "History Shapes" and "Index Map" above!')

  def _check_time(self, raven_vars):
    """
      Checks that "time" is consistent between request and provided variables
      @ In, raven_vars, dict, dictionary of variables from RAVEN
      @ Out, None
    """
    time_var = self._case.get_time_name()
    time_vals = raven_vars.get(time_var, None)
    if self._override_time:
      self._dispatcher.set_time_discr(self._override_time)

    if time_vals is not None:
      req_start, req_end, req_steps = self._dispatcher.get_time_discr()
      # check start time
      if req_start < time_vals[0]:
        raise IOError(f'Requested start time ({req_start}) is less than time variable "{time_var}" ' +
                       f'first value ({time_vals[0]})!')
      # check end time
      if req_end > time_vals[-1]:
        raise IOError(f'Requested end time ({req_end}) is greater than time variable "{time_var}" ' +
                       f'last value ({time_vals[-1]})!')
      # check number of entries
      ## TODO this shouldn't be necessary; we can interpolate!
      ##      for now, though, we don't
      if np.linspace(req_start, req_end, req_steps).size != time_vals.size:
        raise IOError('Requested number of steps ({s}) does not match "{n}" history provided ({g})!'
                       .format(n=time_var,
                               s=np.linspace(req_start, req_end, req_steps).size,
                               g=time_vals.size))

  def _slice_signals(self, all_structure, data):
    """
      Slices from full signals to specific year/cluster/segment
      Target year/cluster/segment are taken from data['active_index'] TODO change this, it's weird
      No indexes from cluster/segment or year should be present after this operation
      @ In, all_structure, dict, dictionary of informations about the source ROM structures
      @ In, data, dict, dictionary of info including RAVEN variables and HERON meta information
      @ Out, truncated, dict, RAVEN_vars portion of "data" but with truncated data
    """
    truncated = {'_indexMap': {}}
    raven = data['RAVEN_vars_full']
    index_map = raven['_indexMap']
    time_var = self._case.get_time_name()

    # are we dealing with time, interpolation, clusters, segments?
    req_indices = data['active_index']
    macro = req_indices['year'] # FIXME Macro ID!
    division = req_indices['division']
    # we keep all of Time, no divider necessary

    for entry in raven:
      if entry in index_map:
        index_order = list(index_map[entry])
        # TODO this is an awkward fix for histories that are missing some of the indexes
        slicer = [np.s_[:]] * len(index_order)
        # if len(index_order) < len(slicer):
        #   slicer = slicer[:len(index_order)]
        # time -> take it all, no action needed
        # cluster
        if '_ROM_Cluster' in index_order:
          cluster_index = index_order.index('_ROM_Cluster')
          slicer[cluster_index] = division
        # macro time (e.g. cycle, year)
        if self._case.get_year_name() in index_order:
          macro_index = index_order.index(self._case.get_year_name())
          slicer[macro_index] = macro
        truncated[entry] = raven[entry][tuple(slicer)]
        truncated['_indexMap'][entry] = [time_var] # the only index left should be "time"
      else:
        # entry doesn't depend on year/cluster, so keep it as is
        truncated[entry] = raven[entry]
    truncated['_indexMap'] = np.atleast_1d(truncated['_indexMap'])
    return truncated


class DispatchManager(ExternalModelPluginBase):
  """
    A plugin to run heron.lib
  """

  def initialize(self, container, runInfoDict, inputFiles):
    """
      Method to initialize the DispatchManager plugin.
      @ In, container, object, external 'self'
      @ In, runInfoDict, dict, the dictionary containing the runInfo (read in the XML input file)
      @ In, inputFiles, list, not used
      @ Out, None
    """
    pass

  def _readMoreXML(self, raven, xml):
    """
      Reads additional inputs for DispatchManager
      @ In, raven, object, variable-storing object
    """
    respec = xml.find('respecTime')
    if respec is not None:
      try:
        stats = [int(x) for x in respec.text.split(',')]
        raven._override_time = stats
        np.linspace(*stats) # test it out
      except Exception:
        raise IOError('DispatchManager xml: respec values should be arguments for np.linspace! Got', respec.text)

  def run(self, raven, raven_dict):
    """
      # TODO split into dispatch manager class and dispatch runner external model
      API for external models.
      This is run as part of the INNER ensemble model, run after the synthetic history generation
      @ In, raven, object, RAVEN variables object
      @ In, raven_dict, dict, additional RAVEN information
      @ Out, None
    """
    path = os.path.join(os.getcwd(), '..', 'heron.lib') # TODO custom name?
    # build runner
    runner = DispatchRunner()
    # load library file
    runner.load_heron_lib(path)
    # load data from RAVEN
    raven_vars = runner.extract_variables(raven, raven_dict)
    # TODO clustering, multiyear, etc?
    # add settings from readMoreXML
    override_time = getattr(raven, '_override_time', None)
    if override_time is not None:
      runner.override_time(override_time) # TODO setter
    dispatch, metrics = runner.run(raven_vars)
    runner.save_variables(raven, dispatch, metrics)






# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the ValuedParam entity.
  These are objects that need to return values, but come from
  a wide variety of different sources.
"""
import os
import sys
import numpy as np

from HERON.src._utils import get_raven_loc
from .ValuedParam import ValuedParam, InputData, InputTypes


class ROM(ValuedParam):
  """
    Represents a ValuedParam that takes values from RAVEN ReducedOrderModel instances.
    The evaluation involves sending in expected inputs and receiving outputs.
  """
  # these types represent values that do not need to be evaluated at run time, as they are determined.
  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a scalar, an ARMA history, or a function
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    # this VP needs sub-VPs for sources of inputs to the ROM
    from .Factory import factory as vp_factory
    spec = InputData.parameterInputFactory('ROM', contentType=InputTypes.StringType,
        descr=r"""indicates that this value will be taken from a RAVEN-trained ROM.""")
    # which ROM is it?
    spec.addParam('rom', param_type=InputTypes.StringType, required=True,
        descr=r"""indicates which DataGenerator ROM should be used for this value.""")
    # which variable(s?) to take from ROM
    spec.addParam('variable', param_type=InputTypes.StringType, required=True,
        descr=r"""indicates the variable output of the ROM from which this value should be taken.""")
    # inputs to the ROM
    descr = r"""designates the source of one of the inputs to the ROM."""
    allowed = ['fixed_value', 'sweep_values', 'opt_bounds', 'variable', 'ARMA', 'Function']
    inp = vp_factory.make_input_specs('input', descr=descr, allowed=allowed)
    descr = r"""name of the variable to pass into the ROM evaluation. Determined by the ROM training."""
    inp.addParam('name', required=True, descr=descr)
    spec.addSub(inp)
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._source_kind = 'ROM'
    self._source_name = None   # name of source ROM
    self._inputs = {}          # map of {name: VP} for all input sources to ROM
    self._output = None        # name of output that should be used

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read valued param from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    # aliases get used to convert variable names, notably for the cashflow's "capacity"
    if alias_dict is None:
      alias_dict = {}
    # which data generator ROM?
    self._source_name = spec.parameterValues['rom']
    # which ROM output to use for value?
    self._output = spec.parameterValues['variable']
    # use VPs for inputs to the ROM
    req_signals = [self._source_name]
    for vp_node in spec.findAll('input'):
      name = vp_node.parameterValues['name']
      new_signals = self.make_sub_vp(name, comp_name, vp_node, mode)
      req_signals.extend(new_signals)
    return req_signals

  def make_sub_vp(self, name, comp, spec, mode):
    """
      Creates a sub VP for this VP
      @ In, name, str, input variable name for ROM
      @ In, comp, str, name of associated component
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ Out, signals, list, variable names needed for evaluation
    """
    from ValuedParamHandler import ValuedParamHandler
    vp = ValuedParamHandler(name)
    signal = vp.read(comp, spec, mode)
    self._inputs[name] = {'vp': vp, 'signals': [signal]}
    return signal

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, run information from RAVEN, including meta and other run info
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, inputs, dict, possibly-modified dictionary of run information
    """
    if aliases is None:
      aliases = {}
    # set the outgoing name for the evaluation results
    key = self._var_name if not target_var else target_var
    # allow aliasing of complex variable names
    var_name = aliases.get(self._output, self._output)
    # the year/cluster indices have already been handled by the Dispatch Manager
    rlz = {}
    ## get the inputs from the input VPs
    for inp_name, inp_info in self._inputs.items():
      vp = inp_info['vp']
      res, _ = vp.evaluate(inputs, target_var=inp_name, aliases=aliases)
      rlz[inp_name] = np.atleast_1d(res[inp_name])
    ## run ROM
    res = self._target_obj.evaluate(rlz)[var_name]
    # NOTE assuming always at least 1d, which is true for RAVEN ROMs I think?
    # TODO can we check this using something more robust?
    if len(res) > 1:
      t = inputs['HERON']['time_index']
      res = res[t]
    else:
      res = res[0]
    return {key: res}, inputs



# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED

from .ExampleValidator import Example
import os
import sys

# set up path of raven
sys.path.append(os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))
import _utils as hutils
try:
  import FARM.src
  farm_loc = FARM.__path__
except ModuleNotFoundError:
  try:
    raven_path = hutils.get_raven_loc()

    # set up path of farm
    farm_loc = hutils.get_farm_loc(raven_path=raven_path)

    if farm_loc is not None:
      farm_path = os.path.abspath(os.path.join(farm_loc))
      sys.path.append(farm_path)
  except ModuleNotFoundError:
    farm_loc = None
  except OSError:
    farm_loc = None

if farm_loc is not None:
  from FARM.src.FARMValidatorsForHeron import FARM_Beta, FARM_Gamma_LTI, FARM_Gamma_FMU, FARM_Delta_FMU

# default known validators
known = {
    'Example': Example,
    # ModelicaGoverner: TODO,
}

# add farm validators to the known dictionary
if farm_loc is not None:
  known['FARM_Beta'] = FARM_Beta
  known['FARM_Gamma_LTI'] = FARM_Gamma_LTI
  known['FARM_Gamma_FMU'] = FARM_Gamma_FMU
  known['FARM_Delta_FMU'] = FARM_Delta_FMU

def get_class(typ):
  """
    Returns the requested dispatcher type.
    @ In, typ, str, name of one of the dispatchers
    @ Out, class, object, class object
  """
  return known.get(typ, None)




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Defines the ValuedParam entity StaticHistory.
  These are objects that need to return values, but come from
  a wide variety of different sources.
"""
from .ValuedParam import ValuedParam, InputData, InputTypes

# class for potentially dynamically-evaluated quantities
class StaticHistory(ValuedParam):
  """
    Represents a ValuedParam that takes values directly from a static CSV file.
  """

  @classmethod
  def get_input_specs(cls):
    """
      Template for parameters that can take a variable from a CSV file.
      @ In, None
      @ Out, spec, InputData, value-based spec
    """
    spec = InputData.parameterInputFactory(
      'CSV',
      contentType=InputTypes.StringType,
      descr=r"""indicates that this value will be taken from a static CSV file signals,
              which will be provided to the dispatch at run time. The value
              of this node should be the name of a static history CSV generator found in the
              \xmlNode{DataGenerator} node."""
    )
    spec.addParam(
      'variable',
      param_type=InputTypes.StringType,
      descr=r"""indicates which variable coming from static history CSV this value should be taken from."""
    )
    return spec

  def __init__(self):
    """
      Constructor.
      @ In, None
      @ Out, None
    """
    super().__init__()
    self._var_name = None # name of the variable within the static hist
    self._source_kind = 'CSV'

  def read(self, comp_name, spec, mode, alias_dict=None):
    """
      Used to read ValuedParam from XML input
      @ In, comp_name, str, name of component that this valued param will be attached to; only used for print messages
      @ In, spec, InputData params, input specifications
      @ In, mode, type of simulation calculation
      @ In, alias_dict, dict, optional, aliases to use for variable naming
      @ Out, needs, list, signals needed to evaluate this ValuedParam at runtime
    """
    super().read(comp_name, spec, mode, alias_dict=None)
    alias_dict = {} if alias_dict is None else alias_dict
    self._source_name = spec.value
    self._var_name = spec.parameterValues['variable']
    return [self._var_name]

  def evaluate(self, inputs, target_var=None, aliases=None):
    """
      Evaluate this ValuedParam, wherever it gets its data from
      @ In, inputs, dict, stuff from RAVEN, particularly including the keys 'meta' and 'raven_vars'
      @ In, target_var, str, optional, requested outgoing variable name if not None
      @ In, aliases, dict, optional, alternate variable names for searching in variables
      @ Out, value, dict, dictionary of resulting evaluation as {vars: vals}
      @ Out, meta, dict, dictionary of meta (possibly changed during evaluation)
    """
    aliases = {} if aliases is None else aliases
    # set the outgoing name for the evaluation results
    key = self._var_name if not target_var else target_var
    # allow aliasing of complex variable names
    var_name = aliases.get(self._var_name, self._var_name)
    # the year/cluster indices have already been handled by the Dispatch Manager
    # -> we just need to know the time index
    t = inputs['HERON']['time_index']
    try:
      value = inputs['HERON']['RAVEN_vars'][var_name][t]
    except KeyError:
      self.raiseAnError(
        RuntimeError,
        f'variable "{var_name}" was not found among the RAVEN variables!' +
        f'Please check to ensure the provided CSV file contains "{var_name}" as a header'
      )
    except IndexError:
      val = inputs['HERON']['RAVEN_vars'][var_name]
      self.raiseAnError(
        RuntimeError,
        f'Attempted to access variable "{var_name}" beyond the end of its length! ' +
        f'Requested index {t} but max index is {len(val)-1}'
      )
    else:
      return {key: value}, inputs




# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Class for mimicking ARMA structure
"""

import numpy as np

def run(raven, raven_dict):
  """
    Makes ARMA-like data structure, same guy every time.
    @ In, raven, object, RAVEN variables object
    @ In, raven_dict, dict, additional RAVEN information
    @ Out, None
  """
  # take some time working hard
  for i in range(int(1e5)):
    x = 2**i
  # result
  raven.NPV = np.random.rand()
  raven.time_delta = 3.14159







# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
  Parses an input file, returning the objects therein.
"""
import sys
import xml.etree.ElementTree as ET

from . import Cases
from . import Components
from . import Placeholders
from . import ValuedParams

from . import _utils as hutils
try:
  import ravenframework
except ModuleNotFoundError:
  raven_path = hutils.get_raven_loc()
  sys.path.append(raven_path)
from ravenframework.utils import xmlUtils



def load(name):
  """
    Loads file into XML format
    @ In, name, str, name of file to read from
    @ Out, xml, ET.element, input as xml
  """
  xml, _ = xmlUtils.loadToTree(name)
  return xml

def parse(xml, loc, messageHandler):
  """
    Read XML and produce object instances
    @ In, xml, ET.element, to read from
    @ In, loc, str, location of input xml
    @ In, messageHandler, RAVEN messageHandler, handler of messages
    @ Out, parse, dict, constructed entities
  """
  # storage
  case = None      # TODO
  components = []  # grid components
  sources = []     # source/demand signals, functions

  # intentionally read case first
  case_node = xml.find('Case')
  if case_node is None:
    raise IOError('<Case> node is missing from HERON input file!')
  case = Cases.Case(loc, messageHandler=messageHandler)
  case.read_input(case_node)

  # components
  components_node = xml.find('Components')
  if components_node is None:
    raise IOError('<Components> node is missing from HERON input file!')
  for comp_xml in components_node:
    comp = Components.Component(messageHandler=messageHandler)
    # check parsing
    comp.read_input(comp_xml, case.get_mode())
    components.append(comp)
  if not components:
    raise IOError('No <Component> nodes were found in the <Components> section!')

  sources_node = xml.find('DataGenerators')
  if sources_node is None:
    raise IOError('<DataGenerators> node is missing from HERON input file!')
  # TODO only load the generators we need?? it'd be a good idea ...
  for sub_xml in sources_node:
    typ = sub_xml.tag
    if typ == 'CSV':
      new = Placeholders.CSV(loc=loc, messageHandler=messageHandler)
    elif typ == 'ARMA':
      new = Placeholders.ARMA(loc=loc, messageHandler=messageHandler)
    elif typ == 'Function':
      new = Placeholders.Function(loc=loc, messageHandler=messageHandler)
    elif typ == 'ROM':
      new = Placeholders.ROM(loc=loc, messageHandler=messageHandler)
    else:
      raise IOError('Unrecognized DataGenerator: "{}"'.format(sub_xml.tag))
    new.read_input(sub_xml)
    sources.append(new)

  # now go back through and link up stuff
  for comp in components:
    found = {}
    for obj, info in comp.get_crossrefs().items():
      found[obj] = {}
      for attr, info in info.items():
        kind, name = info.get_source()
        # if not looking for a DataGenerator placeholder, then nothing more to do
        # if using "activity", also nothing to do
        if kind not in ['Function', 'ARMA', 'ROM', 'CSV']:
          continue
        # find it
        for source in sources:
          if source.is_type(kind) and source.name == name:
            found[obj][attr] = source
            break
        else:
          raise IOError(f'Requested source "{name}" for component "{comp.name}" was not found!')
    comp.set_crossrefs(found)

  # then do pre-writing initialization
  case.initialize(components, sources)

  return {'case': case,
          'components': components,
          'sources': sources,
         }

if __name__ == "__main__":
  inp = load(sys.argv[1])
  obj = parse(inp, None)
  for k, v in obj.items():
    print('  ', k, v)



# Copyright 2020, Battelle Energy Alliance, LLC
# ALL RIGHTS RESERVED
"""
Enables quick testing of Pyomo problem solves, intending to set up a system
similar to how it is set up in the Pyomo dispatcher. This allows rapid testing
of different configurations for the rolling window optimization.
"""
import platform

import numpy as np

import pyomo.environ as pyo
from pyomo.opt import SolverStatus, TerminationCondition

if platform.system() == 'Windows':
  SOLVER = 'glpk'
else:
  SOLVER = 'cbc'

# setup stuff
components = ['steam_source', 'elec_generator', 'steam_storage', 'elec_sink']
resources = ['steam', 'electricity']
time = np.linspace(0, 10, 11) # from @1 to @2 in @3 steps
dt = time[1] - time[0]
resource_map = {'steam_source': {'steam': 0},
                'elec_generator': {'steam': 0, 'electricity': 1},
                'steam_storage': {'steam': 0},
                'elec_sink': {'electricity': 0},
                }
activity = {}
for comp in components:
  activity[comp] = np.zeros((len(resources), len(time)), dtype=float)

# sizing specifications
storage_initial = 50 # kg of steam
storage_limit = 400 # kg of steam
steam_produced = 100 # kg/h of steam
gen_consume_limit = 110 # consumes at most 110 kg/h steam
sink_limit = 10000 # kWh/h = kW of electricity

def make_concrete_model():
  """
    Test writing a simple concrete model with terms typical to the pyomo dispatcher.
    @ In, None
    @ Out, m, pyo.ConcreteModel, instance of the model to solve
  """
  m = pyo.ConcreteModel()
  # indices
  C = np.arange(0, len(components), dtype=int) # indexes component
  R = np.arange(0, len(resources), dtype=int)  # indexes resources
  T = np.arange(0, len(time), dtype=int)       # indexes time
  # move onto model
  m.C = pyo.Set(initialize=C)
  m.R = pyo.Set(initialize=R)
  m.T = pyo.Set(initialize=T)
  # store some stuff for reference -> NOT NOTICED by Pyomo, we hope
  m.Times = time
  m.Components = components
  m.resource_index_map = resource_map
  m.Activity = activity
  #*******************
  #  set up optimization variables
  # -> for now we just do this manually
  # steam_source
  m.steam_source_index_map = pyo.Set(initialize=range(len(m.resource_index_map['steam_source'])))
  m.steam_source_production = pyo.Var(m.steam_source_index_map, m.T, initialize=steam_produced, bounds=(steam_produced, steam_produced))
  # elec_generator
  m.elec_generator_index_map = pyo.Set(initialize=range(len(m.resource_index_map['elec_generator'])))
  m.elec_generator_production = pyo.Var(m.elec_generator_index_map, m.T, initialize=0, bounds=(-gen_consume_limit, 0))
  # steam_storage
  m.steam_storage_index_map = pyo.Set(initialize=range(len(m.resource_index_map['steam_storage'])))
  m.steam_storage_level = pyo.Var(m.steam_storage_index_map, m.T, initialize=0, bounds=(0, storage_limit))
  m.steam_storage_charge = pyo.Var(m.steam_storage_index_map, m.T, initialize=0, bounds=(-storage_limit/dt, 0))
  m.steam_storage_discharge = pyo.Var(m.steam_storage_index_map, m.T, initialize=0, bounds=(0, storage_limit/dt))
  # elec_sink
  m.elec_sink_index_map = pyo.Set(initialize=range(len(m.resource_index_map['elec_sink'])))
  m.elec_sink_production = pyo.Var(m.elec_sink_index_map, m.T, initialize=0, bounds=(-sink_limit, 0))
  #*******************
  #  set up lower, upper bounds
  # -> for now we just do this manually
  # -> consuming is negative sign by convention!
  # -> producing is positive sign by convention!
  # steam source produces exactly 100 steam
  #    m.steam_source_lower_limit = pyo.Constraint(m.T, rule=lambda m, t: m.steam_source_production[0, t] >= steam_produced)
  #    m.steam_source_upper_limit = pyo.Constraint(m.T, rule=lambda m, t: m.steam_source_production[0, t] <= steam_produced)
  # elec generator can consume steam to produce electricity; 0 < consumed steam < 1000
  # -> this effectively limits electricity production, but we're defining capacity in steam terms for fun
  # -> therefore signs are negative, -1000 < consumed steam < 0!
  #    m.elec_generator_lower_limit = pyo.Constraint(m.T, rule=lambda m, t: m.elec_generator_production[0, t] >= -gen_consume_limit)
  #    m.elec_generator_upper_limit = pyo.Constraint(m.T, rule=lambda m, t: m.elec_generator_production[0, t] <= 0)
  # elec sink can take any amount of electricity
  # -> consuming, so -10000 < consumed elec < 0
  #    m.elec_sink_lower_limit = pyo.Constraint(m.T, rule=lambda m, t: m.elec_sink_production[0, t] >= -sink_limit)
  #    m.elec_sink_upper_limit = pyo.Constraint(m.T, rule=lambda m, t: m.elec_sink_production[0, t] <= 0)
  # storage is in LEVEL not ACTIVITY (e.g. kg not kg/s) -> lets say it can store X kg
  #    m.steam_storage_lower_limit = pyo.Constraint(m.T, rule=lambda m, t: m.steam_storage_level[0, t] >= 0)
  #    m.steam_storage_upper_limit = pyo.Constraint(m.T, rule=lambda m, t: m.steam_storage_level[0, t] <= storage_limit)
  #*******************
  # create transfer function
  # 2 steam make 1 electricity (sure, why not)
  m.elec_generator_transfer = pyo.Constraint(m.T, rule=_generator_transfer)
  #*******************
  # create conservation rules
  # steam
  m.steam_conservation = pyo.Constraint(m.T, rule=_conserve_steam)
  # electricity
  m.elec_conservation = pyo.Constraint(m.T, rule=_conserve_electricity)
  #*******************
  # create objective function
  m.OBJ = pyo.Objective(sense=pyo.maximize, rule=_economics)
  #######
  # return
  return m

#######
#
# Callback Functions
#
def _generator_transfer(m, t):
  """
    Constraint rule for electricity generation in generator
    @ In, m, pyo.ConcreteModel, model containing problem
    @ In, t, int, time indexer
    @ Out, constraint, bool, constraining evaluation
  """
  return - m.elec_generator_production[0, t] == 2.0 * m.elec_generator_production[1, t]

def _conserve_steam(m, t):
  """
    Constraint rule for conserving steam
    @ In, m, pyo.ConcreteModel, model containing problem
    @ In, t, int, time indexer
    @ Out, constraint, bool, constraining evaluation
  """
  # signs are tricky here, consumption is negative and production is positive
  # a positive delta in steam storage level means it absorbed steam, so it's a negative term
  storage_source = - (m.steam_storage_level[0, t] - (storage_initial if t == 0 else (m.steam_storage_charge[0, t-1] + m.steam_storage_discharge[0, t-1]))) / dt
  sources = storage_source + m.steam_source_production[0, t]
  sinks = m.elec_generator_production[0, t]
  return sources + sinks == 0

def _conserve_electricity(m, t):
  """
    Constraint rule for conserving electricity
    @ In, m, pyo.ConcreteModel, model containing problem
    @ In, t, int, time indexer
    @ Out, constraint, bool, constraining evaluation
  """
  sources = m.elec_generator_production[1, t]
  sinks = m.elec_sink_production[0, t]
  return sources + sinks == 0

def _economics(m):
  """
    Constraint rule for optimization target
    @ In, m, pyo.ConcreteModel, model containing problem
    @ Out, objective, float, constraining evaluation
  """
  opex = sum(m.elec_generator_production[0, t] for t in m.T) * 10 # will be negative b/c consumed
  sales = - sum((m.elec_sink_production[0, t] * (100 if t < 5 else 1)) for t in m.T) # net positive because consumed
  return opex + sales

#######
#
# Debug printing functions
#
def print_setup(m):
  """
    Debug printing for pre-solve model setup
    @ In, m, pyo.ConcreteModel, model containing problem
    @ Out, None
  """
  print('/' + '='*80)
  print('DEBUGG model pieces:')
  print('  -> objective:')
  print('     ', m.OBJ.pprint())
  print('  -> variables:')
  for var in m.component_objects(pyo.Var):
    print('     ', var.pprint())
  print('  -> constraints:')
  for constr in m.component_objects(pyo.Constraint):
    print('     ', constr.pprint())
  print('\\' + '='*80)
  print('')

def print_solution(m):
  """
    Debug printing for post-solve model setup
    @ In, m, pyo.ConcreteModel, model containing problem
    @ Out, None
  """
  print('')
  print('*'*80)
  print('solution:')
  print('  objective value:', m.OBJ())
  print('time | steam source | storage level | elec gen (s, e) | elec sink')
  for t in m.T:
    print(f'{m.Times[t]:1.2e} | ' +
        f'{m.steam_source_production[0, t].value: 1.3e} | ' +
        f'{m.steam_storage_level[0, t].value: 1.3e} | ' +
        f'({m.elec_generator_production[0, t].value: 1.3e}, {m.elec_generator_production[1, t].value: 1.3e}) | ' +
        f'{m.elec_sink_production[0, t].value: 1.3e}'
        )
  print('*'*80)

#######
#
# Solver.
#
def solve_model(m):
  """
    Solves the model.
    @ In, m, pyo.ConcreteModel, model containing problem
    @ Out, m, pyo.ConcreteModel, results
  """
  soln = pyo.SolverFactory(SOLVER).solve(m)
  return soln

if __name__ == '__main__':
  m = make_concrete_model()
  print_setup(m)
  s = solve_model(m)
  print_solution(m)

# solution using setup:
#   time = np.linspace(0, 10, 11)
#   storage_initial = 50 # kg of steam
#   storage_limit = 400 # kg of steam
#   steam_produced = 100 # kg/h of steam
#   gen_consume_limit = 110 # consumes at most 110 kg/h steam
#   sink_limit = 10000 # kWh/h = kW of electricity
#   1 steam = 2 * electricity
#   cost for generator = 10 * kg/h steam consumed
#   profit = 100 * electricity consumed at sink, t < 5,
#              1 * electricity consumed at sink, t >= 5,
#
# should look like:
# ********************************************************************************
# solution:
#   objective value: 20100.0
# time | steam source | steam storage | elec gen (s, e) | elec sink
# 0.00e+00 |  1.000e+02 |  4.000e+01 | (-1.100e+02,  5.500e+01) | -5.500e+01
# 1.00e+00 |  1.000e+02 |  3.000e+01 | (-1.100e+02,  5.500e+01) | -5.500e+01
# 2.00e+00 |  1.000e+02 |  2.000e+01 | (-1.100e+02,  5.500e+01) | -5.500e+01
# 3.00e+00 |  1.000e+02 |  1.000e+01 | (-1.100e+02,  5.500e+01) | -5.500e+01
# 4.00e+00 |  1.000e+02 |  0.000e+00 | (-1.100e+02,  5.500e+01) | -5.500e+01
# 5.00e+00 |  1.000e+02 |  0.000e+00 | (-1.000e+02,  5.000e+01) | -5.000e+01
# 6.00e+00 |  1.000e+02 |  0.000e+00 | (-1.000e+02,  5.000e+01) | -5.000e+01
# 7.00e+00 |  1.000e+02 |  1.000e+02 | ( 0.000e+00,  0.000e+00) |  0.000e+00
# 8.00e+00 |  1.000e+02 |  2.000e+02 | ( 0.000e+00,  0.000e+00) |  0.000e+00
# 9.00e+00 |  1.000e+02 |  3.000e+02 | ( 0.000e+00,  0.000e+00) |  0.000e+00
# 1.00e+01 |  1.000e+02 |  4.000e+02 | ( 0.000e+00,  0.000e+00) |  0.000e+00
# ********************************************************************************
